# Copyright 2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Logic dealing with bases."""

from typing import Tuple, Union

from charmcraft.models.charmcraft import Base
from charmcraft.utils import get_host_architecture, get_os_platform


def get_host_as_base() -> Base:
    """Get host OS represented as Base.

    The host OS name is translated to lower-case for consistency.

    :returns: Base configuration matching host.
    """
    os_platform = get_os_platform()
    host_arch = get_host_architecture()
    name = os_platform.system.lower()
    channel = os_platform.release

    return Base(name=name, channel=channel, architectures=[host_arch])


def check_if_base_matches_host(base: Base) -> Tuple[bool, Union[str, None]]:
    """Check if given base matches the host.

    :param base: Base to check.

    :returns: Tuple of bool indicating whether it is a match, with optional
              reason if not a match.
    """
    host_base = get_host_as_base()
    host_arch = host_base.architectures[0]

    if host_base.name != base.name:
        return False, f"name {base.name!r} does not match host {host_base.name!r}"

    # For Ubuntu, MacOS and Windows, use the full version.
    # For other OSes, use the major version only.

    if host_base.name in ("ubuntu", "darwin", "windows"):
        host_channel = host_base.channel
    else:
        host_channel = host_base.channel.split(".")[0]
    if host_channel != base.channel:
        return (
            False,
            f"channel {base.channel!r} does not match host {host_base.channel!r}",
        )

    if host_arch not in base.architectures:
        return (
            False,
            f"host architecture {host_arch!r} not in base architectures {base.architectures!r}",
        )

    return True, None



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import re
from textwrap import dedent

import pydantic
import pytest
from craft_cli import CraftError

from charmcraft.metafiles.metadata import parse_charm_metadata_yaml, read_metadata_yaml

# tests for parsing metadata


def test_parse_metadata_yaml_complete(tmp_path):
    """Example of parsing with all the optional attributes."""
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text(
        """
        name: test-name
        summary: Test summary
        description: Lot of text.
    """
    )

    metadata = parse_charm_metadata_yaml(tmp_path)

    assert metadata.name == "test-name"
    assert metadata.summary == "Test summary"
    assert metadata.description == "Lot of text."


@pytest.mark.parametrize(
    "metadata_yaml_template",
    [
        dedent(
            """\
            name: {name}
            summary: Test summary
            description: Lot of text.
            """
        ),
    ],
)
@pytest.mark.parametrize("name", ["name1", "my-charm-foo"])
def test_parse_metadata_yaml_valid_names(
    tmp_path, name, prepare_metadata_yaml, metadata_yaml_template
):
    prepare_metadata_yaml(metadata_yaml_template.format(name=name))

    metadata = parse_charm_metadata_yaml(tmp_path)

    assert metadata.name == name


@pytest.mark.parametrize(
    "metadata_yaml_template",
    [
        dedent(
            """\
            name: {name}
            summary: Test summary
            description: Lot of text.
            """
        ),
    ],
)
@pytest.mark.parametrize("name", [1, "false", "[]"])
def test_parse_metadata_yaml_error_invalid_names(
    tmp_path, prepare_metadata_yaml, metadata_yaml_template, name
):
    prepare_metadata_yaml(metadata_yaml_template.format(name=name))

    with pytest.raises(pydantic.ValidationError):
        parse_charm_metadata_yaml(tmp_path)


def test_parse_metadata_yaml_error_missing(tmp_path):
    msg = re.escape(
        "Cannot read the metadata.yaml file: FileNotFoundError(2, 'No such file or directory')"
    )
    with pytest.raises(CraftError, match=msg):
        parse_charm_metadata_yaml(tmp_path)


# tests for reading metadata raw content


def test_read_metadata_yaml_complete(tmp_path):
    """Example of parsing with all the optional attributes."""
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text(
        """
        name: test-name
        summary: Test summary
        description: Text.
    """
    )

    metadata = read_metadata_yaml(tmp_path)
    assert metadata == {"name": "test-name", "summary": "Test summary", "description": "Text."}


def test_read_metadata_yaml_error_invalid(tmp_path):
    """Open a metadata.yaml that would fail verification."""
    metadata_file = tmp_path / "metadata.yaml"
    metadata_file.write_text("- whatever")
    metadata = read_metadata_yaml(tmp_path)
    assert metadata == ["whatever"]


def test_read_metadata_yaml_error_missing(tmp_path):
    """Do not hide the file not being accessible."""
    with pytest.raises(OSError):
        read_metadata_yaml(tmp_path)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Submodule for handling the all extra yaml files."""

import pathlib
from typing import Any, Dict

import yaml

from craft_cli import emit

__all__ = ["actions", "manifest", "metadata", "config", "read_yaml"]


def read_yaml(yaml_file_path: pathlib.Path) -> Dict[str, Any]:
    """Parse yaml file.

    :returns: the YAML decoded yaml content
    """
    emit.debug(f"Reading {str(yaml_file_path)!r}")
    with yaml_file_path.open("rt", encoding="utf8") as fh:
        return yaml.safe_load(fh)



# Copyright 2020-2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Expose needed names at main package level."""

from importlib.metadata import version, PackageNotFoundError
import os


def _get_version() -> str:
    if os.getenv("SNAP_NAME") == "charmcraft":
        return os.getenv("SNAP_VERSION", "")
    try:
        return version("charmcraft")
    except PackageNotFoundError:
        return "devel"


__version__ = _get_version()



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft configuration pydantic model."""
import datetime
import os
import pathlib
from typing import Any, Dict, List, Literal, Optional, Union

import pydantic
from craft_cli import CraftError

from charmcraft.const import (
    CHARM_METADATA_KEYS,
    CHARM_METADATA_LEGACY_KEYS,
    METADATA_FILENAME,
)
from charmcraft.extensions import apply_extensions
from charmcraft.format import format_pydantic_errors
from charmcraft.metafiles.actions import parse_actions_yaml
from charmcraft.metafiles.config import parse_config_yaml
from charmcraft.metafiles.metadata import (
    parse_bundle_metadata_yaml,
    parse_charm_metadata_yaml,
)
from charmcraft.models.actions import JujuActions
from charmcraft.models.basic import AttributeName, LinterName, ModelConfigDefaults
from charmcraft.models.config import JujuConfig
from charmcraft.parts import process_part_config
from charmcraft.utils import get_host_architecture


class CharmhubConfig(
    ModelConfigDefaults,
    alias_generator=lambda s: s.replace("_", "-"),
):
    """Definition of Charmhub endpoint configuration."""

    api_url: pydantic.HttpUrl = "https://api.charmhub.io"
    storage_url: pydantic.HttpUrl = "https://storage.snapcraftcontent.com"
    registry_url: pydantic.HttpUrl = "https://registry.jujucharms.com"


class Base(ModelConfigDefaults):
    """Represents a base."""

    name: pydantic.StrictStr
    channel: pydantic.StrictStr
    architectures: List[pydantic.StrictStr] = [get_host_architecture()]


class BasesConfiguration(
    ModelConfigDefaults,
    alias_generator=lambda s: s.replace("_", "-"),
):
    """Definition of build-on/run-on combinations."""

    build_on: List[Base]
    run_on: List[Base]


class Project(ModelConfigDefaults):
    """Internal-only project configuration."""

    # do not verify that `dirpath` is a valid existing directory; it's used externally as a dir
    # to load the config itself (so we're really do the validation there), and we want to support
    # the case of a missing directory (and still load a default config structure)
    dirpath: pathlib.Path
    config_provided: bool = False

    # this timestamp will be used in several places, even sent to Charmhub: needs to be UTC
    started_at: datetime.datetime


class Ignore(ModelConfigDefaults):
    """Definition of `analysis.ignore` configuration."""

    attributes: List[AttributeName] = []
    linters: List[LinterName] = []


class AnalysisConfig(ModelConfigDefaults, allow_population_by_field_name=True):
    """Definition of `analysis` configuration."""

    ignore: Ignore = Ignore()


class Links(ModelConfigDefaults):
    """Definition of `links` in metadata."""

    contact: Optional[Union[pydantic.StrictStr, List[pydantic.StrictStr]]]
    documentation: Optional[pydantic.AnyHttpUrl]
    issues: Optional[Union[pydantic.AnyHttpUrl, List[pydantic.AnyHttpUrl]]]
    source: Optional[Union[pydantic.AnyHttpUrl, List[pydantic.AnyHttpUrl]]]
    website: Optional[Union[pydantic.AnyHttpUrl, List[pydantic.AnyHttpUrl]]]


class CharmcraftConfig(
    ModelConfigDefaults,
    validate_all=False,
    alias_generator=lambda s: s.replace("_", "-"),
):
    """Definition of charmcraft.yaml configuration."""

    # this needs to go before 'parts', as it used by the validator
    project: Project

    metadata_legacy: bool = False

    type: Literal["bundle", "charm"]
    name: Optional[pydantic.StrictStr]
    summary: Optional[pydantic.StrictStr]
    description: Optional[pydantic.StrictStr]
    charmhub: CharmhubConfig = CharmhubConfig()
    parts: Optional[Dict[str, Any]]
    bases: Optional[List[BasesConfiguration]]
    analysis: AnalysisConfig = AnalysisConfig()
    actions: Optional[JujuActions]
    assumes: Optional[List[Union[str, Dict[str, Union[List, Dict]]]]]
    containers: Optional[Dict[str, Any]]
    devices: Optional[Dict[str, Any]]
    title: Optional[pydantic.StrictStr]
    extra_bindings: Optional[Dict[str, Any]]
    peers: Optional[Dict[str, Any]]
    provides: Optional[Dict[str, Any]]
    requires: Optional[Dict[str, Any]]
    resources: Optional[Dict[str, Any]]
    storage: Optional[Dict[str, Any]]
    subordinate: Optional[bool]
    terms: Optional[List[str]]
    links: Optional[Links]
    config: Optional[JujuConfig]

    @pydantic.validator("name", pre=True, always=True)
    def validate_name(cls, name, values):
        """Verify charm name is valid with exception when instantiated without YAML."""
        if values.get("type") == "charm" and not name:
            raise ValueError("needs value")

        return name

    @pydantic.validator("summary", pre=True, always=True)
    def validate_summary(cls, summary, values):
        """Verify charm summary is valid with exception when instantiated without YAML."""
        if values.get("type") == "charm" and not summary:
            raise ValueError("needs value")

        return summary

    @pydantic.validator("description", pre=True, always=True)
    def validate_description(cls, description, values):
        """Verify charm name is valid with exception when instantiated without YAML."""
        if values.get("type") == "charm" and not description:
            raise ValueError("needs value")

        return description

    @pydantic.validator("parts", pre=True, always=True)
    def validate_special_parts(cls, parts, values):
        """Verify parts type (craft-parts will re-validate the schemas for the plugins)."""
        if "type" not in values:
            # we need 'type' to be set in this validator; if not there it's an error in
            # the schema anyway, so the whole loading will fail (no need to raise an
            # extra error here, it gets confusing to the user)
            return None

        if parts is None:
            # no parts indicated, default to the type of package
            parts = {values["type"]: {}}

        if not isinstance(parts, dict):
            raise TypeError("value must be a dictionary")

        for name, part in parts.items():
            if not isinstance(part, dict):
                raise TypeError(f"part {name!r} must be a dictionary")
            # implicit plugin fixup
            if "plugin" not in part:
                part["plugin"] = name

        # if needed, create 'source' properties for special parts "charm" with plugin "charm".
        # and "bundle" with plugin "bundle", pointing to project's directory
        for name, part in parts.items():
            if name == "charm" and part["plugin"] == "charm":
                part.setdefault("source", str(values["project"].dirpath))

            if name == "bundle" and part["plugin"] == "bundle":
                part.setdefault("source", str(values["project"].dirpath))

        return parts

    @pydantic.validator("parts", each_item=True)
    def validate_each_part(cls, item):
        """Verify each part in the parts section. Craft-parts will re-validate them."""
        return process_part_config(item)

    @pydantic.validator("bases", pre=True)
    def validate_bases_presence(cls, bases, values):
        """Forbid 'bases' in bundles.

        This is to avoid a possible confusion of expecting the bundle
        to be built in a specific environment
        """
        if values.get("type") == "bundle":
            raise ValueError("Field not allowed when type=bundle")
        return bases

    @pydantic.validator("actions", pre=True, always=True)
    def validate_actions(cls, actions, values):
        """Verify 'actions' in charms.

        Currently, actions will be passed through to the charms.
        And individual "actions.yaml" should not exists when actions
        is defined in charmcraft.yaml.
        """
        actions_yaml = parse_actions_yaml(values["project"].dirpath, allow_broken=True)
        if actions is None:
            return actions_yaml
        else:
            if actions_yaml is not None:
                raise ValueError(
                    "'actions.yaml' file not allowed when an 'actions' section is "
                    "defined in 'charmcraft.yaml'"
                )

            return JujuActions.parse_obj({"actions": actions})

    @pydantic.validator("config", pre=True, always=True)
    def validate_config(cls, config, values):
        """Verify 'actions' in charms.

        Currently, actions will be passed through to the charms.
        And individual "actions.yaml" should not exists when actions
        is defined in charmcraft.yaml.
        """
        config_yaml = parse_config_yaml(values["project"].dirpath, allow_broken=True)
        if config is None:
            return config_yaml
        else:
            if config_yaml is not None:
                raise ValueError(
                    "'config.yaml' file not allowed when an 'config' section is "
                    "defined in 'charmcraft.yaml'"
                )

            return JujuConfig.parse_obj(config)

    @classmethod
    def expand_short_form_bases(cls, bases: List[Dict[str, Any]]) -> None:
        """Expand short-form base configuration into long-form in-place."""
        for index, base in enumerate(bases):
            # Skip if already long-form. Account for common typos in case user
            # intends to use long-form, but did so incorrectly (for better
            # error message handling).
            if "run-on" in base or "run_on" in base or "build-on" in base or "build_on" in base:
                continue

            try:
                converted_base = Base(**base)
            except pydantic.ValidationError as error:
                # Rewrite location to assist user.
                pydantic_errors = error.errors()
                for pydantic_error in pydantic_errors:
                    pydantic_error["loc"] = ("bases", index, pydantic_error["loc"][0])

                raise CraftError(format_pydantic_errors(pydantic_errors))

            base.clear()
            base["build-on"] = [converted_base.dict()]
            base["run-on"] = [converted_base.dict()]

    @classmethod
    def unmarshal(cls, obj: Dict[str, Any], project: Project):
        """Unmarshal object with necessary translations and error handling.

        (1) Perform any necessary translations.

        (2) Standardize error reporting.

        :returns: valid CharmcraftConfig.

        :raises CraftError: On failure to unmarshal object.
        """
        try:
            # Expand short-form bases if only the bases is a valid list. If it
            # is not a valid list, parse_obj() will properly handle the error.
            if isinstance(obj.get("bases"), list):
                cls.expand_short_form_bases(obj["bases"])

            obj = apply_extensions(project.dirpath, obj)

            # Re-expand it in case extensions added short-form bases.
            if isinstance(obj.get("bases"), list):
                cls.expand_short_form_bases(obj["bases"])

            # If metadata.yaml exists, try merge it into config.
            if os.path.isfile(project.dirpath / METADATA_FILENAME):
                # metadata.yaml exists, so we can't specify metadata keys in charmcraft.yaml.
                for key in CHARM_METADATA_KEYS.union(CHARM_METADATA_LEGACY_KEYS):
                    if key in obj:
                        raise CraftError(
                            f"Cannot specify '{key}' in charmcraft.yaml when "
                            f"'{METADATA_FILENAME}' exists"
                        )

                if obj.get("type") == "charm":
                    metadata_legacy = parse_charm_metadata_yaml(project.dirpath, allow_basic=True)

                    # need to copy 3 fields from metadata_legacy to charmcraft config
                    return cls.parse_obj(
                        {
                            "project": project,
                            "name": metadata_legacy.name,
                            "summary": metadata_legacy.summary,
                            "description": metadata_legacy.description,
                            "metadata-legacy": True,
                            **obj,
                        }
                    )
                elif obj.get("type") == "bundle":
                    # bundle may not have metadata.yaml.
                    # but if it does, it should have name and optional description
                    # metadata.yaml will be copied without validation if it exists
                    metadata_legacy = parse_bundle_metadata_yaml(project.dirpath)
                    return cls.parse_obj(
                        {
                            "project": project,
                            "name": metadata_legacy.name,
                            "description": metadata_legacy.description,
                            "metadata-legacy": True,
                            **obj,
                        }
                    )
                else:
                    # fallthrough for pydantic to handle
                    pass

            return cls.parse_obj({"project": project, **obj})
        except pydantic.ValidationError as error:
            raise CraftError(format_pydantic_errors(error.errors()))

    @classmethod
    def schema(cls, **kwargs) -> Dict[str, Any]:
        """Perform any schema fixups required to hide internal details."""
        schema = super().schema(**kwargs)

        # The internal __root__ detail is leaked, overwrite it.
        schema["properties"]["parts"]["default"] = {}

        # Project is an internal detail, purge references.
        schema["definitions"].pop("Project", None)
        schema["properties"].pop("project", None)
        schema["required"].remove("project")
        return schema



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import itertools
import os
import re
import subprocess
import sys

import pytest

from charmcraft import __version__, main


def get_python_filepaths(*, roots=None, python_paths=None):
    """Helper to retrieve paths of Python files."""
    if python_paths is None:
        python_paths = ["setup.py"]
    if roots is None:
        roots = ["charmcraft", "tests"]
    for root in roots:
        for dirpath, dirnames, filenames in os.walk(root):
            for filename in filenames:
                if filename.endswith(".py"):
                    python_paths.append(os.path.join(dirpath, filename))
    return python_paths


def test_ensure_copyright():
    """Check that all non-empty Python files have copyright somewhere in the first 5 lines."""
    issues = []
    regex = re.compile(r"# Copyright \d{4}(-\d{4})? Canonical Ltd.$")
    for filepath in get_python_filepaths():
        if os.stat(filepath).st_size == 0:
            continue

        with open(filepath, encoding="utf8") as fh:
            for line in itertools.islice(fh, 5):
                if regex.match(line):
                    break
            else:
                issues.append(filepath)
    if issues:
        msg = "Please add copyright headers to the following files:\n" + "\n".join(issues)
        pytest.fail(msg, pytrace=False)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_setup_version():
    """Verify that setup.py is picking up the version correctly."""
    cmd = [os.path.abspath("setup.py"), "--version"]
    proc = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
    output = proc.stdout.decode("utf8")
    assert output.strip() == __version__


def test_bashcompletion_all_commands():
    """Verify that all commands are represented in the bash completion file."""
    # get the line where all commands are specified in the completion file; this is custom
    # to our file, but simple and good enough
    completed_commands = None
    with open("completion.bash", encoding="utf8") as fh:
        completion_text = fh.read()
    m = re.search(r"cmds=\((.*?)\)", completion_text, re.DOTALL)
    if m:
        completed_commands = set(m.groups()[0].split())
    else:
        pytest.fail("Failed to find commands in the bash completion file")

    real_command_names = set()
    for cgroup in main.COMMAND_GROUPS:
        real_command_names.update(cmd.name for cmd in cgroup.commands if not cmd.hidden)

    assert completed_commands == real_command_names



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft metadata pydantic model."""

from typing import Any, Dict, List, Optional, Union

import pydantic
from craft_cli import CraftError

from charmcraft.const import METADATA_FILENAME


class CharmMetadataLegacy(
    pydantic.BaseModel,
    extra=pydantic.Extra.allow,
    frozen=True,
    validate_all=True,
    alias_generator=lambda s: s.replace("_", "-"),
):
    """Object representing LEGACY charm metadata.yaml contents.

    This model only supports the legacy charm metadata.yaml format for compatibility.
    New metadata defined in charmcraft.yaml is handled by the CharmcraftConfig model.

    specs: https://juju.is/docs/sdk/metadata-yaml
    """

    name: pydantic.StrictStr
    summary: pydantic.StrictStr
    description: pydantic.StrictStr
    assumes: Optional[List[Union[str, Dict[str, Union[List, Dict]]]]]
    containers: Optional[Dict[str, Any]]
    devices: Optional[Dict[str, Any]]
    display_name: Optional[pydantic.StrictStr]
    docs: Optional[pydantic.AnyHttpUrl]
    extra_bindings: Optional[Dict[str, Any]]
    issues: Optional[Union[pydantic.AnyHttpUrl, List[pydantic.AnyHttpUrl]]]
    maintainers: Optional[List[pydantic.StrictStr]]
    peers: Optional[Dict[str, Any]]
    provides: Optional[Dict[str, Any]]
    requires: Optional[Dict[str, Any]]
    resources: Optional[Dict[str, Any]]
    source: Optional[Union[pydantic.AnyHttpUrl, List[pydantic.AnyHttpUrl]]]
    storage: Optional[Dict[str, Any]]
    subordinate: Optional[bool]
    terms: Optional[List[pydantic.StrictStr]]
    website: Optional[Union[pydantic.AnyHttpUrl, List[pydantic.AnyHttpUrl]]]

    @classmethod
    def unmarshal(cls, obj: Dict[str, Any]):
        """Unmarshal object with necessary translations and error handling.

        :returns: valid CharmMetadataLegacy object.

        :raises CraftError: On failure to unmarshal object.
        """
        # convert undocumented "maintainer" to documented "maintainers"
        if "maintainer" in obj and "maintainers" in obj:
            raise CraftError(
                f"Cannot specify both 'maintainer' and 'maintainers' in {METADATA_FILENAME}"
            )

        if "maintainer" in obj:
            obj["maintainers"] = [obj["maintainer"]]
            del obj["maintainer"]

        return cls.parse_obj(obj)


class BundleMetadataLegacy(
    pydantic.BaseModel,
    extra=pydantic.Extra.allow,
    frozen=True,
    validate_all=True,
    alias_generator=lambda s: s.replace("_", "-"),
):
    """Object representing LEGACY bundle metadata.yaml contents.

    This model only supports the legacy bundle metadata.yaml format.

    specs: https://juju.is/docs/sdk/metadata-yaml
    """

    name: pydantic.StrictStr
    description: Optional[pydantic.StrictStr]

    @classmethod
    def unmarshal(cls, obj: Dict[str, Any]):
        """Unmarshal object with necessary translations and error handling.

        :returns: valid BundleMetadataLegacy.

        :raises CraftError: On failure to unmarshal object.
        """
        return cls.parse_obj(obj)



# Copyright 2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import pathlib
import sys
from subprocess import CalledProcessError, CompletedProcess
from unittest.mock import call, patch

import craft_parts
import pydantic
import pytest
import pytest_subprocess
from craft_parts import plugins
from craft_parts.errors import PluginEnvironmentValidationError

from charmcraft.parts import reactive

pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")


@pytest.fixture()
def charm_exe(tmp_path):
    """Provide a fake charm executable."""
    charm_bin = pathlib.Path(tmp_path, "mock_bin", "charm")
    charm_bin.parent.mkdir(exist_ok=True)
    charm_bin.write_text(
        '#!/bin/sh\necho \'{"charm-tools": {"version": "2.8.4", "git": "+git-7-6126e17", '
        '"gitn": 7, "gitsha": "6126e17", "pre_release": false, "snap": "+snap-x12"}}\''
    )
    charm_bin.chmod(0o755)
    return charm_bin


@pytest.fixture()
def broken_charm_exe(tmp_path):
    """Provide a fake charm executable that fails to run."""
    charm_bin = pathlib.Path(tmp_path, "mock_bin", "charm")
    charm_bin.parent.mkdir(exist_ok=True)
    charm_bin.write_text('#!/bin/sh\nexit 1"')
    charm_bin.chmod(0o755)
    return charm_bin


@pytest.fixture()
def spec(tmp_path):
    """Provide a common spec to build the different artifacts."""
    return {
        "plugin": "reactive",
        "source": str(tmp_path),
        "reactive-charm-build-arguments": [
            "--charm-argument",
            "--charm-argument-with argument",
        ],
    }


@pytest.fixture()
def plugin_properties(spec):
    return reactive.ReactivePluginProperties.unmarshal(spec)


@pytest.fixture()
def plugin(tmp_path, plugin_properties, spec):
    project_dirs = craft_parts.ProjectDirs(work_dir=tmp_path)
    part_spec = plugins.extract_part_properties(spec, plugin_name="reactive")
    part = craft_parts.Part(
        "foo", part_spec, project_dirs=project_dirs, plugin_properties=plugin_properties
    )
    project_info = craft_parts.ProjectInfo(
        application_name="test",
        project_dirs=project_dirs,
        cache_dir=tmp_path,
        project_name="fake-project",
    )
    part_info = craft_parts.PartInfo(project_info=project_info, part=part)

    return plugins.get_plugin(part=part, part_info=part_info, properties=plugin_properties)


def test_get_build_package(plugin):
    assert plugin.get_build_packages() == set()


def test_get_build_snaps(plugin):
    assert plugin.get_build_snaps() == set()


def test_get_build_environment(plugin):
    assert plugin.get_build_environment() == {}


def test_get_build_commands(plugin, tmp_path):
    assert plugin.get_build_commands() == [
        f"{sys.executable} -I {reactive.__file__} fake-project "
        f"{tmp_path}/parts/foo/build {tmp_path}/parts/foo/install "
        "--charm-argument --charm-argument-with argument"
    ]


def test_invalid_properties(plugin):
    with pytest.raises(pydantic.ValidationError) as raised:
        plugin.properties_class.unmarshal({"source": ".", "reactive-invalid": True})
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("reactive-invalid",)
    assert err[0]["type"] == "value_error.extra"


def test_validate_environment(plugin, plugin_properties, charm_exe):
    validator = plugin.validator_class(
        part_name="my-part",
        env=f"PATH={str(charm_exe.parent)}",
        properties=plugin_properties,
    )
    validator.validate_environment()


def test_validate_environment_with_charm_part(plugin, plugin_properties):
    validator = plugin.validator_class(
        part_name="my-part", env="PATH=/foo", properties=plugin_properties
    )
    validator.validate_environment(part_dependencies=["charm-tools"])


def test_validate_missing_charm(
    fake_process: pytest_subprocess.FakeProcess, plugin, plugin_properties
):
    fake_process.register(["/bin/bash", fake_process.any()], returncode=127)
    validator = plugin.validator_class(
        part_name="my-part", env="/foo", properties=plugin_properties
    )
    with pytest.raises(PluginEnvironmentValidationError) as raised:
        validator.validate_environment()

    assert raised.value.reason == (
        "charm tool not found and part 'my-part' does not depend on a part named 'charm-tools'"
    )


def test_validate_broken_charm(plugin, plugin_properties, broken_charm_exe):
    validator = plugin.validator_class(
        part_name="my-part",
        env=f"PATH={str(broken_charm_exe.parent)}",
        properties=plugin_properties,
    )
    with pytest.raises(PluginEnvironmentValidationError) as raised:
        validator.validate_environment()

    assert raised.value.reason == "charm tools failed with error code 2"


@pytest.fixture()
def build_dir(tmp_path):
    build_dir = tmp_path / "build"
    build_dir.mkdir()

    return build_dir


@pytest.fixture()
def install_dir(tmp_path):
    install_dir = tmp_path / "install"
    install_dir.mkdir()

    return install_dir


@pytest.fixture()
def fake_run():
    patcher = patch("subprocess.run")
    yield patcher.start()
    patcher.stop()


def test_build(build_dir, install_dir, fake_run):
    fake_run.return_value = CompletedProcess(("charm", "build"), 0)
    returncode = reactive.build(
        charm_name="test-charm",
        build_dir=build_dir,
        install_dir=install_dir,
        charm_build_arguments=["--charm-argument", "--charm-argument-with", "argument"],
    )

    assert returncode == 0
    assert not (build_dir / "test-charm").exists()
    assert fake_run.mock_calls == [
        call(["charm", "proof"], check=True),
        call(
            [
                "charm",
                "build",
                "--charm-argument",
                "--charm-argument-with",
                "argument",
                "-o",
                str(build_dir),
            ],
            check=True,
        ),
    ]


def test_build_charm_proof_raises_error_messages(build_dir, install_dir, fake_run):
    fake_run.side_effect = CalledProcessError(200, "E: name missing")

    returncode = reactive.build(
        charm_name="test-charm",
        build_dir=build_dir,
        install_dir=install_dir,
        charm_build_arguments=["--charm-argument", "--charm-argument-with", "argument"],
    )

    assert returncode == 200
    assert not (build_dir / "test-charm").exists()
    assert fake_run.mock_calls == [
        call(["charm", "proof"], check=True),
    ]


def test_build_charm_proof_raises_warning_messages_does_not_raise(
    build_dir, install_dir, fake_run
):
    fake_run.side_effect = CalledProcessError(100, "W: Description is not pretty")

    returncode = reactive.build(
        charm_name="test-charm",
        build_dir=build_dir,
        install_dir=install_dir,
        charm_build_arguments=["--charm-argument", "--charm-argument-with", "argument"],
    )

    assert returncode == 0
    assert not (build_dir / "test-charm").exists()
    assert fake_run.mock_calls == [
        call(["charm", "proof"], check=True),
        call(
            [
                "charm",
                "build",
                "--charm-argument",
                "--charm-argument-with",
                "argument",
                "-o",
                str(build_dir),
            ],
            check=True,
        ),
    ]


def test_build_charm_build_raises_error_messages(build_dir, install_dir, fake_run):
    def _run_generator():
        """Generate return values and exceptions alternatively.

        This cannot be done directly because passing an iterable to `side_effect` pivots the
        mocks return_value, and does not allow us to raise an actual exception.
        """
        yield CompletedProcess(("charm", "proof"), 0)
        yield CalledProcessError(200, "E: name missing")
        yield CompletedProcess(("charm", "proof"), 0)
        yield CalledProcessError(-1, "E: name missing")

    fake_run.side_effect = _run_generator()

    returncode = reactive.build(
        charm_name="test-charm",
        build_dir=build_dir,
        install_dir=install_dir,
        charm_build_arguments=["--charm-argument", "--charm-argument-with", "argument"],
    )

    assert returncode == 200
    assert not (build_dir / "test-charm").exists()
    assert fake_run.mock_calls == [
        call(["charm", "proof"], check=True),
        call(
            [
                "charm",
                "build",
                "--charm-argument",
                "--charm-argument-with",
                "argument",
                "-o",
                str(build_dir),
            ],
            check=True,
        ),
    ]

    # Also ensure negative return codes raises error
    returncode = reactive.build(
        charm_name="test-charm",
        build_dir=build_dir,
        install_dir=install_dir,
        charm_build_arguments=["--charm-argument", "--charm-argument-with", "argument"],
    )
    assert returncode == -1


def test_build_charm_build_raises_warning_messages_does_not_raise(
    build_dir, install_dir, fake_run
):
    def _run_generator():
        """Generate return values and exceptions alternatively.

        This cannot be done directly because passing an iterable to `side_effect` pivots the
        mocks return_value, and does not allow us to raise an actual exception.
        """
        yield CompletedProcess(("charm", "proof"), 0)
        yield CalledProcessError(100, "W: Description is not pretty")

    fake_run.side_effect = _run_generator()

    returncode = reactive.build(
        charm_name="test-charm",
        build_dir=build_dir,
        install_dir=install_dir,
        charm_build_arguments=["--charm-argument", "--charm-argument-with", "argument"],
    )

    assert returncode == 0
    assert not (build_dir / "test-charm").exists()
    assert fake_run.mock_calls == [
        call(["charm", "proof"], check=True),
        call(
            [
                "charm",
                "build",
                "--charm-argument",
                "--charm-argument-with",
                "argument",
                "-o",
                str(build_dir),
            ],
            check=True,
        ),
    ]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import sys

import pytest

from charmcraft.utils.yaml import load_yaml


def test_load_yaml_success(tmp_path):
    test_file = tmp_path / "testfile.yaml"
    test_file.write_text(
        """
        foo: 33
    """
    )
    content = load_yaml(test_file)
    assert content == {"foo": 33}


def test_load_yaml_no_file(tmp_path, emitter):
    test_file = tmp_path / "testfile.yaml"
    content = load_yaml(test_file)
    assert content is None

    expected = f"Couldn't find config file {str(test_file)!r}"
    emitter.assert_debug(expected)


def test_load_yaml_directory(tmp_path, emitter):
    test_file = tmp_path / "testfile.yaml"
    test_file.mkdir()
    content = load_yaml(test_file)
    assert content is None

    expected = f"Couldn't find config file {str(test_file)!r}"
    emitter.assert_debug(expected)


def test_load_yaml_corrupted_format(tmp_path, emitter):
    test_file = tmp_path / "testfile.yaml"
    test_file.write_text(
        """
        foo: [1, 2
    """
    )
    content = load_yaml(test_file)
    assert content is None

    expected = "Failed to read/parse config file.*testfile.yaml.*ParserError.*"
    emitter.assert_debug(expected, regex=True)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_load_yaml_file_problem(tmp_path, emitter):
    test_file = tmp_path / "testfile.yaml"
    test_file.write_text(
        """
        foo: bar
    """
    )
    test_file.chmod(0o000)
    content = load_yaml(test_file)
    assert content is None

    expected = f"Failed to read/parse config file {str(test_file)!r}.*PermissionError.*"
    emitter.assert_debug(expected, regex=True)



# Copyright 2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import contextlib
import pathlib
import re
import sys
from unittest.mock import Mock, call, patch

import pytest
from craft_cli import CraftError
from craft_providers import ProviderError, bases, lxd, multipass
from craft_providers.actions.snap_installer import Snap

from charmcraft import providers
from charmcraft.models.charmcraft import Base, BasesConfiguration
from charmcraft.snap import CharmcraftSnapConfiguration


@pytest.fixture()
def mock_provider(mock_instance, fake_provider):
    mock_provider = Mock(wraps=fake_provider)
    with patch("charmcraft.providers.get_provider", return_value=mock_provider):
        yield mock_provider


@pytest.fixture()
def mock_is_base_available():
    with patch(
        "charmcraft.providers.is_base_available", return_value=(True, None)
    ) as mock_is_base_available:
        yield mock_is_base_available


@pytest.fixture()
def mock_check_if_base_matches_host():
    with patch(
        "charmcraft.providers.check_if_base_matches_host", return_value=(True, None)
    ) as mock_check_if_base_matches_host:
        yield mock_check_if_base_matches_host


@pytest.fixture()
def mock_get_host_architecture():
    with patch(
        "charmcraft.providers.get_host_architecture", return_value="host-arch"
    ) as mock_arch:
        yield mock_arch


@pytest.fixture()
def mock_snap_config():
    with patch("charmcraft.providers.get_snap_configuration", return_value=None) as mock_snap:
        yield mock_snap


@pytest.fixture()
def mock_is_developer_mode():
    with patch(
        "charmcraft.providers.is_charmcraft_running_in_developer_mode",
        return_value=False,
    ) as mock_is_dev_mode:
        yield mock_is_dev_mode


@pytest.fixture()
def mock_is_snap():
    with patch(
        "charmcraft.providers.is_charmcraft_running_from_snap", return_value=False
    ) as mock_is_snap:
        yield mock_is_snap


@pytest.fixture()
def simple_base_config():
    """Yields a simple BaseConfiguration object."""
    return [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                ],
                "run-on": [
                    Base(name="x2name", channel="x2channel", architectures=["x2arch"]),
                ],
            }
        ),
    ]


@pytest.fixture()
def complex_base_config():
    """Yields a complex list of BaseConfiguration objects."""
    return [
        # 1 build-on and 1 run-on
        BasesConfiguration(
            **{
                "build-on": [
                    Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                ],
                "run-on": [
                    Base(name="x2name", channel="x2channel", architectures=["x2arch"]),
                ],
            }
        ),
        # 2 build-on and 1 run-on
        BasesConfiguration(
            **{
                "build-on": [
                    Base(name="x3name", channel="x3channel", architectures=["x3arch"]),
                    Base(name="x4name", channel="x4channel", architectures=["x4arch"]),
                ],
                "run-on": [
                    Base(name="x5name", channel="x5channel", architectures=["x5arch"]),
                ],
            }
        ),
        # 1 build-on and 2 run-on with multiple architectures
        BasesConfiguration(
            **{
                "build-on": [
                    Base(name="x6name", channel="x6channel", architectures=["x6arch"]),
                ],
                "run-on": [
                    Base(name="x7name", channel="x7channel", architectures=["x7arch"]),
                    Base(
                        name="x8name",
                        channel="x8channel",
                        architectures=["x8arch1", "x8arch2"],
                    ),
                ],
            }
        ),
    ]


def test_create_build_plan_simple(
    emitter, mock_provider, mock_is_base_available, simple_base_config
):
    """Verify creation of a simple build plan."""
    build_plan = providers.create_build_plan(
        bases=simple_base_config,
        bases_indices=None,
        destructive_mode=False,
        managed_mode=False,
        provider=mock_provider,
    )

    assert build_plan == [
        providers.Plan(
            bases_config=BasesConfiguration(
                **{
                    "build-on": [
                        Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                    ],
                    "run-on": [
                        Base(name="x2name", channel="x2channel", architectures=["x2arch"]),
                    ],
                }
            ),
            build_on=Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
            bases_index=0,
            build_on_index=0,
        ),
    ]
    emitter.assert_interactions(
        [
            call("debug", "Building for 'bases[0]' as host matches 'build-on[0]'."),
        ]
    )


def test_create_build_plan_complex(
    emitter, complex_base_config, mock_provider, mock_is_base_available
):
    """Verify creation of a complex build plan."""

    build_plan = providers.create_build_plan(
        bases=complex_base_config,
        bases_indices=None,
        destructive_mode=False,
        managed_mode=False,
        provider=mock_provider,
    )

    assert build_plan == [
        providers.Plan(
            bases_config=BasesConfiguration(
                **{
                    "build-on": [
                        Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                    ],
                    "run-on": [
                        Base(name="x2name", channel="x2channel", architectures=["x2arch"]),
                    ],
                }
            ),
            build_on=Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
            bases_index=0,
            build_on_index=0,
        ),
        providers.Plan(
            bases_config=BasesConfiguration(
                **{
                    "build-on": [
                        Base(name="x3name", channel="x3channel", architectures=["x3arch"]),
                        Base(name="x4name", channel="x4channel", architectures=["x4arch"]),
                    ],
                    "run-on": [
                        Base(name="x5name", channel="x5channel", architectures=["x5arch"]),
                    ],
                }
            ),
            build_on=Base(name="x3name", channel="x3channel", architectures=["x3arch"]),
            bases_index=1,
            build_on_index=0,
        ),
        providers.Plan(
            bases_config=BasesConfiguration(
                **{
                    "build-on": [
                        Base(name="x6name", channel="x6channel", architectures=["x6arch"]),
                    ],
                    "run-on": [
                        Base(name="x7name", channel="x7channel", architectures=["x7arch"]),
                        Base(
                            name="x8name",
                            channel="x8channel",
                            architectures=["x8arch1", "x8arch2"],
                        ),
                    ],
                }
            ),
            build_on=Base(name="x6name", channel="x6channel", architectures=["x6arch"]),
            bases_index=2,
            build_on_index=0,
        ),
    ]
    emitter.assert_interactions(
        [
            call("debug", "Building for 'bases[0]' as host matches 'build-on[0]'."),
            call("debug", "Building for 'bases[1]' as host matches 'build-on[0]'."),
            call("debug", "Building for 'bases[2]' as host matches 'build-on[0]'."),
        ]
    )


@pytest.mark.parametrize(("destructive_mode", "managed_mode"), [(True, False), (False, True)])
def test_create_build_plan_base_matches_host(
    emitter,
    destructive_mode,
    managed_mode,
    mock_check_if_base_matches_host,
    mock_provider,
    simple_base_config,
):
    """Verify the first `build_on` Base that matches the host is used for the build plan
    when building in managed mode or destructive mode."""
    build_plan = providers.create_build_plan(
        bases=simple_base_config,
        bases_indices=None,
        destructive_mode=destructive_mode,
        managed_mode=managed_mode,
        provider=mock_provider,
    )

    assert build_plan == [
        providers.Plan(
            bases_config=BasesConfiguration(
                **{
                    "build-on": [
                        Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                    ],
                    "run-on": [
                        Base(name="x2name", channel="x2channel", architectures=["x2arch"]),
                    ],
                }
            ),
            build_on=Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
            bases_index=0,
            build_on_index=0,
        ),
    ]
    emitter.assert_interactions(
        [
            call("debug", "Building for 'bases[0]' as host matches 'build-on[0]'."),
        ]
    )


def test_create_build_plan_is_base_available(emitter, mock_is_base_available, mock_provider):
    """Verify the first available `build_on` Base that is used for the build plan."""
    base = [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                    Base(name="x2name", channel="x2channel", architectures=["x2arch"]),
                ],
                "run-on": [
                    Base(name="x3name", channel="x3channel", architectures=["x3arch"]),
                ],
            }
        )
    ]

    # the first Base is not available, but the second Base is available
    mock_is_base_available.side_effect = [(False, "test error message"), (True, None)]

    build_plan = providers.create_build_plan(
        bases=base,
        bases_indices=None,
        destructive_mode=False,
        managed_mode=False,
        provider=mock_provider,
    )

    # verify charmcraft will build on the second Base
    assert build_plan[0].build_on == Base(
        name="x2name", channel="x2channel", architectures=["x2arch"]
    )
    assert build_plan[0].build_on_index == 1

    emitter.assert_interactions(
        [
            call("progress", "Skipping 'bases[0].build-on[0]': test error message."),
            call("debug", "Building for 'bases[0]' as host matches 'build-on[1]'."),
        ]
    )


def test_create_build_plan_base_index_usage(
    complex_base_config,
    emitter,
    mock_is_base_available,
    mock_provider,
):
    """Verify `bases_indices` argument causes build plan to only contain matching bases."""
    build_plan = providers.create_build_plan(
        bases=complex_base_config,
        bases_indices=[1, 2],
        destructive_mode=False,
        managed_mode=False,
        provider=mock_provider,
    )

    assert build_plan == [
        providers.Plan(
            bases_config=BasesConfiguration(
                **{
                    "build-on": [
                        Base(name="x3name", channel="x3channel", architectures=["x3arch"]),
                        Base(name="x4name", channel="x4channel", architectures=["x4arch"]),
                    ],
                    "run-on": [
                        Base(name="x5name", channel="x5channel", architectures=["x5arch"]),
                    ],
                }
            ),
            build_on=Base(name="x3name", channel="x3channel", architectures=["x3arch"]),
            bases_index=1,
            build_on_index=0,
        ),
        providers.Plan(
            bases_config=BasesConfiguration(
                **{
                    "build-on": [
                        Base(name="x6name", channel="x6channel", architectures=["x6arch"]),
                    ],
                    "run-on": [
                        Base(name="x7name", channel="x7channel", architectures=["x7arch"]),
                        Base(
                            name="x8name",
                            channel="x8channel",
                            architectures=["x8arch1", "x8arch2"],
                        ),
                    ],
                }
            ),
            build_on=Base(name="x6name", channel="x6channel", architectures=["x6arch"]),
            bases_index=2,
            build_on_index=0,
        ),
    ]

    emitter.assert_interactions(
        [
            call("debug", "Skipping 'bases[0]' due to --base-index usage."),
            call("debug", "Building for 'bases[1]' as host matches 'build-on[0]'."),
            call("debug", "Building for 'bases[2]' as host matches 'build-on[0]'."),
        ]
    )


def test_create_build_plan_no_suitable_bases(
    emitter, complex_base_config, mock_is_base_available, mock_provider
):
    """Verify an empty build plan is returned when no bases are available."""
    mock_is_base_available.return_value = (False, "test error message")

    build_plan = providers.create_build_plan(
        bases=complex_base_config,
        bases_indices=None,
        destructive_mode=False,
        managed_mode=False,
        provider=mock_provider,
    )

    assert build_plan == []

    emitter.assert_interactions(
        [
            call("progress", "Skipping 'bases[0].build-on[0]': test error message."),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[0]' configuration.",
                permanent=True,
            ),
            call("progress", "Skipping 'bases[1].build-on[0]': test error message."),
            call("progress", "Skipping 'bases[1].build-on[1]': test error message."),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[1]' configuration.",
                permanent=True,
            ),
            call("progress", "Skipping 'bases[2].build-on[0]': test error message."),
            call(
                "progress",
                "No suitable 'build-on' environment found in 'bases[2]' configuration.",
                permanent=True,
            ),
        ]
    )


def test_create_build_plan_no_bases_error(mock_provider):
    """Verify an error is raised when no bases are passed."""
    with pytest.raises(CraftError) as error:
        providers.create_build_plan(
            bases=None,
            bases_indices=None,
            destructive_mode=False,
            managed_mode=False,
            provider=mock_provider,
        )

    assert str(error.value) == "Cannot create build plan because no bases were provided."


def test_get_command_environment_minimal(monkeypatch):
    monkeypatch.setenv("IGNORE_ME", "or-im-failing")
    monkeypatch.setenv("PATH", "not-using-host-path")

    env = providers.get_command_environment(bases.ubuntu.BuilddBase)

    assert env == {
        "CHARMCRAFT_MANAGED_MODE": "1",
        "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/snap/bin",
    }


def test_get_command_environment_all_opts(monkeypatch):
    monkeypatch.setenv("IGNORE_ME", "or-im-failing")
    monkeypatch.setenv("PATH", "not-using-host-path")
    monkeypatch.setenv("http_proxy", "test-http-proxy")
    monkeypatch.setenv("https_proxy", "test-https-proxy")
    monkeypatch.setenv("no_proxy", "test-no-proxy")

    env = providers.get_command_environment(bases.ubuntu.BuilddBase)

    assert env == {
        "CHARMCRAFT_MANAGED_MODE": "1",
        "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/snap/bin",
        "http_proxy": "test-http-proxy",
        "https_proxy": "test-https-proxy",
        "no_proxy": "test-no-proxy",
    }


@pytest.mark.parametrize(
    ("bases_index", "build_on_index", "project_name", "target_arch", "expected"),
    [
        (0, 0, "mycharm", "test-arch1", "charmcraft-mycharm-{inode}-0-0-test-arch1"),
        (
            1,
            2,
            "my-other-charm",
            "test-arch2",
            "charmcraft-my-other-charm-{inode}-1-2-test-arch2",
        ),
    ],
)
def test_get_instance_name(
    bases_index, build_on_index, project_name, target_arch, expected, tmp_path
):
    assert providers.get_instance_name(
        bases_index=bases_index,
        build_on_index=build_on_index,
        project_name=project_name,
        project_path=tmp_path,
        target_arch=target_arch,
    ) == expected.format(inode=tmp_path.stat().st_ino)


@pytest.mark.parametrize(
    "cache_path",
    [
        None,
        pathlib.Path("charmcraft-cache"),
    ],
)
@pytest.mark.parametrize(
    ("platform", "snap_channel", "expected_snap_channel"),
    [
        ("linux", None, None),
        ("linux", "edge", "edge"),
        ("darwin", "edge", "edge"),
        # default to stable on non-linux system
        ("darwin", None, "stable"),
    ],
)
@pytest.mark.parametrize(
    "alias_ubuntu",
    [
        bases.ubuntu.BuilddBaseAlias.BIONIC,
        bases.ubuntu.BuilddBaseAlias.FOCAL,
        bases.ubuntu.BuilddBaseAlias.JAMMY,
    ],
)
def test_get_base_configuration_ubuntu(
    platform,
    snap_channel,
    expected_snap_channel,
    alias_ubuntu,
    cache_path,
    mocker,
    tmp_path,
):
    """Verify the snapcraft snap is installed from the correct channel."""
    mocker.patch("sys.platform", platform)
    mocker.patch(
        "charmcraft.providers.get_managed_environment_snap_channel",
        return_value=snap_channel,
    )
    mocker.patch("charmcraft.providers.get_command_environment", return_value="test-env")
    mocker.patch("charmcraft.providers.get_instance_name", return_value="test-instance-name")
    mock_buildd_base = mocker.patch("craft_providers.bases.ubuntu.BuilddBase")
    mock_buildd_base.compatibility_tag = "buildd-base-v0"

    if cache_path:
        cache_path = tmp_path / cache_path
        cache_path.mkdir(parents=True, exist_ok=True)

    providers.get_base_configuration(
        alias=alias_ubuntu, instance_name="test-instance-name", shared_cache_path=cache_path
    )

    mock_buildd_base.assert_called_with(
        alias=alias_ubuntu,
        environment="test-env",
        hostname="test-instance-name",
        snaps=[Snap(name="charmcraft", channel=expected_snap_channel, classic=True)],
        compatibility_tag="charmcraft-buildd-base-v0.0",
        cache_path=cache_path,
    )


@pytest.mark.parametrize(
    "cache_path",
    [
        None,
        pathlib.Path("charmcraft-cache"),
    ],
)
@pytest.mark.parametrize(
    ("platform", "snap_channel", "expected_snap_channel"),
    [
        ("linux", None, None),
        ("linux", "edge", "edge"),
        ("darwin", "edge", "edge"),
        # default to stable on non-linux system
        ("darwin", None, "stable"),
    ],
)
def test_get_base_configuration_centos(
    platform,
    snap_channel,
    expected_snap_channel,
    cache_path,
    mocker,
    tmp_path,
):
    """Verify the snapcraft snap is installed from the correct channel."""
    mocker.patch("sys.platform", platform)
    mocker.patch(
        "charmcraft.providers.get_managed_environment_snap_channel",
        return_value=snap_channel,
    )
    mocker.patch("charmcraft.providers.get_command_environment", return_value="test-env")
    mocker.patch("charmcraft.providers.get_instance_name", return_value="test-instance-name")
    mock_centos_base = mocker.patch("craft_providers.bases.centos.CentOSBase")
    mock_centos_base.compatibility_tag = "centos-base-v0"

    if cache_path:
        cache_path = tmp_path / cache_path
        cache_path.mkdir(parents=True, exist_ok=True)

    providers.get_base_configuration(
        alias=bases.centos.CentOSBaseAlias.SEVEN,
        instance_name="test-instance-name",
        shared_cache_path=cache_path,
    )

    mock_centos_base.assert_called_with(
        alias=bases.centos.CentOSBaseAlias.SEVEN,
        environment="test-env",
        hostname="test-instance-name",
        snaps=[Snap(name="charmcraft", channel=expected_snap_channel, classic=True)],
        compatibility_tag="charmcraft-centos-base-v0.0",
        cache_path=cache_path,
    )


def test_capture_logs_from_instance_ok(emitter, mock_instance, tmp_path, mocker):
    @contextlib.contextmanager
    def fake_pull(source, missing_ok):
        assert source == pathlib.Path("/tmp/charmcraft.log")
        assert missing_ok is True
        fake_file = tmp_path / "fake.file"
        fake_file.write_text("some\nlog data\nhere")
        yield fake_file

    mocker.patch.object(mock_instance, "temporarily_pull_file", fake_pull)
    providers.capture_logs_from_instance(mock_instance)

    emitter.assert_interactions(
        [
            call("debug", "Logs captured from managed instance:"),
            call("debug", ":: some"),
            call("debug", ":: log data"),
            call("debug", ":: here"),
        ]
    )


def test_capture_logs_from_instance_not_found(emitter, mock_instance, tmp_path, mocker):
    @contextlib.contextmanager
    def fake_pull(source, missing_ok):
        yield None  # didn't find the indicated file

    mocker.patch.object(mock_instance, "temporarily_pull_file", fake_pull)
    providers.capture_logs_from_instance(mock_instance)

    emitter.assert_debug("No logs found in instance.")


def test_ensure_provider_is_available_installed_yes(mocker, fake_provider):
    """Verify provider is ensured to be available when installed (fake_provider's default)."""
    confirmation_mock = mocker.patch("charmcraft.providers.confirm_with_user")
    available_mock = mocker.patch.object(fake_provider, "ensure_provider_is_available")

    providers.ensure_provider_is_available(fake_provider)

    confirmation_mock.assert_not_called()
    available_mock.assert_called_once()


def test_ensure_provider_is_available_installed_no_user_confirms_yes(mocker, fake_provider):
    """Verify provider is ensured to be available, not installed but user chooses to install."""
    confirmation_mock = mocker.patch("charmcraft.providers.confirm_with_user", return_value=True)
    mocker.patch.object(fake_provider, "is_provider_installed", return_value=False)
    available_mock = mocker.patch.object(fake_provider, "ensure_provider_is_available")

    providers.ensure_provider_is_available(fake_provider)

    message = (
        "Provider is required but not installed. Do you wish to "
        "install provider and configure it with the defaults?"
    )
    confirmation_mock.assert_called_with(message, default=False)
    available_mock.assert_called_once()


def test_ensure_provider_is_available_installed_no_user_confirms_no(mocker, fake_provider):
    """Raise an error if not installed and the user does not choose to install it."""
    mocker.patch("charmcraft.providers.confirm_with_user", return_value=False)
    mocker.patch.object(fake_provider, "is_provider_installed", return_value=False)

    with pytest.raises(ProviderError) as error:
        providers.ensure_provider_is_available(fake_provider)

    assert error.value.brief == "Provider is required, but not installed."


@pytest.mark.parametrize(
    ("name", "channel", "architectures", "expected_valid", "expected_reason"),
    [
        ("ubuntu", "18.04", ["host-arch"], True, None),
        ("ubuntu", "20.04", ["host-arch"], True, None),
        ("ubuntu", "22.04", ["host-arch"], True, None),
        ("ubuntu", "20.04", ["extra-arch", "host-arch"], True, None),
        (
            "not-ubuntu",
            "20.04",
            ["host-arch"],
            False,
            r"name 'not-ubuntu' is not yet supported \(must be 'ubuntu', .*\)",
        ),
        (
            "ubuntu",
            "10.04",
            ["host-arch"],
            False,
            r"base 'ubuntu' channel '10.04' is not yet supported \(must be .*\)",
        ),
        (
            "ubuntu",
            "20.04",
            ["other-arch"],
            False,
            r"host architecture 'host-arch' not in base architectures \['other-arch'\]",
        ),
    ],
)
def test_is_base_available(
    mock_get_host_architecture, name, channel, architectures, expected_valid, expected_reason
):
    base = Base(name=name, channel=channel, architectures=architectures)
    valid, reason = providers.is_base_available(base)

    assert valid == expected_valid
    if reason is None:
        assert expected_reason == reason
    else:
        assert re.fullmatch(expected_reason, reason)


def test_get_provider_default(mock_snap_config, mock_is_developer_mode, mock_is_snap):
    if sys.platform == "linux":
        provider = providers.get_provider()
        assert isinstance(provider, lxd.LXDProvider)
        assert provider.lxd_project == "charmcraft"
    else:
        assert isinstance(providers.get_provider(), multipass.MultipassProvider)


def test_get_provider_developer_mode_env(
    mock_snap_config, mock_is_developer_mode, mock_is_snap, monkeypatch
):
    mock_is_developer_mode.return_value = True
    monkeypatch.setenv("CHARMCRAFT_PROVIDER", "lxd")
    provider = providers.get_provider()
    assert isinstance(provider, lxd.LXDProvider)
    assert provider.lxd_project == "charmcraft"

    monkeypatch.setenv("CHARMCRAFT_PROVIDER", "multipass")
    assert isinstance(providers.get_provider(), multipass.MultipassProvider)


def test_get_provider_snap_config(mock_is_snap, mock_is_developer_mode, mock_snap_config):
    mock_is_snap.return_value = True

    mock_snap_config.return_value = CharmcraftSnapConfiguration(provider="lxd")
    provider = providers.get_provider()
    assert isinstance(provider, lxd.LXDProvider)
    assert provider.lxd_project == "charmcraft"

    mock_snap_config.return_value = CharmcraftSnapConfiguration(provider="multipass")
    assert isinstance(providers.get_provider(), multipass.MultipassProvider)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charm project related utilities."""
import itertools
import os
import pathlib
import sys
from collections import defaultdict
from typing import Container, Dict

from craft_cli import emit
from jinja2 import Environment, FileSystemLoader, PackageLoader, StrictUndefined

from charmcraft.errors import DuplicateCharmsError, InvalidCharmPathError
from charmcraft.utils.yaml import load_yaml


def find_charm_sources(
    base_path: pathlib.Path, charm_names: Container[str]
) -> Dict[str, pathlib.Path]:
    """Find all charm directories matching the given names under a base path.

    :param base_path: The base directory under which to look.
    :param charm_names: A container with charm names to find.
    :returns: A dictionary mapping charm names to their paths.
    :raises: DuplicateCharmsError if a charm is found in multiple directories.
    """
    duplicate_charms = defaultdict(list)
    charm_paths: Dict[str, pathlib.Path] = {}
    outer_potential_paths = itertools.chain(
        (p.parent.resolve() for p in base_path.glob("charms/*/metadata.yaml")),
        (p.parent.resolve() for p in base_path.glob("operators/*/metadata.yaml")),
        (p.parent.resolve() for p in base_path.glob("*/metadata.yaml")),
    )
    potential_paths = filter(lambda p: (p / "charmcraft.yaml").exists(), outer_potential_paths)
    for path in potential_paths:
        if path in charm_paths.values():  # Symlinks can cause ignorable duplicate paths.
            continue
        try:
            charm_name = get_charm_name_from_path(path)
        except InvalidCharmPathError:
            continue
        if charm_name not in charm_names:  # We only care if the charm is listed for finding
            continue
        if charm_name != path.name:
            emit.verbose(f"Charm {charm_name!r} found in non-matching path {path}")
            continue
        if charm_name in charm_paths:
            duplicate_charms[charm_name].append(path)
        else:
            charm_paths[charm_name] = path
    if duplicate_charms:
        raise DuplicateCharmsError(duplicate_charms)
    return charm_paths


def get_charm_name_from_path(path: pathlib.Path) -> str:
    """Get a charm's name from a given path.

    :param path: The path to investigate.
    :returns: The name of the charm in this path
    :raises: InvalidCharmPathError if the path given is not a valid charm source.
    """
    charmcraft_yaml = load_yaml(path / "charmcraft.yaml")
    if charmcraft_yaml is None or charmcraft_yaml.get("type") != "charm":
        raise InvalidCharmPathError(path)
    metadata_yaml = load_yaml(path / "metadata.yaml")
    if metadata_yaml is None or "name" not in metadata_yaml:
        raise InvalidCharmPathError(path)
    return metadata_yaml["name"]


def get_templates_environment(templates_dir):
    """Create and return a Jinja environment to deal with the templates."""
    templates_dir = os.path.join("templates", templates_dir)
    if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        # Running as PyInstaller bundle. For more information:
        # https://pyinstaller.readthedocs.io/en/stable/runtime-information.html
        # In this scenario we need to load from the data location that is unpacked
        # into the temporary directory at runtime (sys._MEIPASS).
        emit.debug(f"Bundle directory: {sys._MEIPASS}")
        loader = FileSystemLoader(os.path.join(sys._MEIPASS, templates_dir))
    else:
        loader = PackageLoader("charmcraft", templates_dir)

    return Environment(
        loader=loader,
        autoescape=False,  # no need to escape things here :-)
        keep_trailing_newline=True,  # they're not text files if they don't end in newline!
        optimized=False,  # optimization doesn't make sense for one-offs
        undefined=StrictUndefined,
    )  # fail on undefined



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import argparse
import ast
import itertools
import os
import subprocess
import sys
from argparse import ArgumentParser
from textwrap import dedent
from unittest.mock import patch

import pytest
from craft_cli import (
    ArgumentParsingError,
    CommandGroup,
    CraftError,
    EmitterMode,
    ProvideHelpException,
)
from craft_store.errors import CraftStoreError

from charmcraft import __version__, env, utils
from charmcraft.cmdbase import FORMAT_HELP_STR, JSON_FORMAT, BaseCommand
from charmcraft.commands.store.client import ALTERNATE_AUTH_ENV_VAR
from charmcraft.main import COMMAND_GROUPS, _get_system_details, main

# --- Tests for the main entry point

# In all the test methods below we patch Dispatcher.run so we don't really exercise any
# command machinery, even if we call to main using a real command (which is to just
# make argument parsing system happy).


def test_main_ok():
    """Work ended ok: message handler notified properly, return code in 0."""
    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.return_value = None
            retcode = main(["charmcraft", "version"])

    assert retcode == 0
    emit_mock.ended_ok.assert_called_once_with()

    # check how Emitter was initted
    emit_mock.init.assert_called_once_with(
        EmitterMode.BRIEF,
        "charmcraft",
        f"Starting charmcraft version {__version__}",
        log_filepath=None,
    )


def test_main_managed_instance_init(monkeypatch):
    """Init emitter with a specific log filepath."""
    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")

    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.return_value = None
            main(["charmcraft", "version"])

    # check how Emitter was initted
    emit_mock.init.assert_called_once_with(
        EmitterMode.BRIEF,
        "charmcraft",
        f"Starting charmcraft version {__version__}",
        log_filepath=env.get_managed_environment_log_path(),
    )


@pytest.mark.parametrize(
    "side_effect",
    [
        ValueError("broken"),
        KeyboardInterrupt("interrupted"),
        CraftStoreError("bad server"),
        CraftError("bad charmcraft"),
    ],
)
def test_main_managed_instance_error(monkeypatch, side_effect, config):
    """The managed instance will not expose the "internal" log filepath."""
    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")

    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.pre_parse_args") as d_mock:
            d_mock.side_effect = side_effect
            main(["charmcraft", "version"])

    # check that the error sent to Craft CLI will not report the logpath
    (end_in_error_call,) = emit_mock.error.mock_calls
    error = end_in_error_call.args[0]
    assert error.logpath_report is False


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_main_load_config_ok(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Command is properly executed, after loading and receiving the config."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    class MyCommand(BaseCommand):
        help_msg = "some help"
        name = "cmdname"
        overview = "test overview"

        def run(self, parsed_args):
            assert self.config.type == "bundle"

    with patch("charmcraft.main.COMMAND_GROUPS", [CommandGroup("title", [MyCommand])]):
        retcode = main(["charmcraft", "cmdname", f"--project-dir={tmp_path}"])
    assert retcode == 0


def test_main_load_config_not_present_ok():
    """Config is not present but the command does not need it."""

    class MyCommand(BaseCommand):
        help_msg = "some help"
        name = "cmdname"
        overview = "test overview"

        def run(self, parsed_args):
            assert not self.config.project.config_provided
            self._check_config()

    with patch("charmcraft.main.COMMAND_GROUPS", [CommandGroup("title", [MyCommand])]):
        retcode = main(["charmcraft", "cmdname", "--project-dir=/whatever"])
    assert retcode == 0


def test_main_load_config_not_present_but_needed(capsys):
    """Config is not present and the command needs it."""

    class MyCommand(BaseCommand):
        help_msg = "some help"
        name = "cmdname"
        overview = "test overview"

        def run(self, parsed_args):
            assert not self.config.project.config_provided
            self._check_config(config_file=True)

    with patch("charmcraft.main.COMMAND_GROUPS", [CommandGroup("title", [MyCommand])]):
        retcode = main(["charmcraft", "cmdname", "--project-dir=/whatever"])
    assert retcode == 1

    out, err = capsys.readouterr()
    assert not out
    assert err == (
        "The specified command needs a valid 'charmcraft.yaml' configuration file (in "
        "the current directory or where specified with --project-dir option); see "
        "the reference: https://discourse.charmhub.io/t/charmcraft-configuration/4138\n"
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_main_load_config_bases_not_present_but_not_needed(
    capsys,
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """Config bases is not present and the command does not need it."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    class MyCommand(BaseCommand):
        help_msg = "some help"
        name = "cmdname"
        overview = "test overview"

        def run(self, parsed_args):
            self._check_config(config_file=True, bases=False)
            assert self.config.type == "charm"

    with patch("charmcraft.main.COMMAND_GROUPS", [CommandGroup("title", [MyCommand])]):
        retcode = main(["charmcraft", "cmdname", f"--project-dir={tmp_path}"])
    assert retcode == 0


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_main_load_config_bases_not_present_but_needed(
    capsys,
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """Config bases is not present and the command needs it."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    class MyCommand(BaseCommand):
        help_msg = "some help"
        name = "cmdname"
        overview = "test overview"

        def run(self, parsed_args):
            assert self.config.type == "charm"
            self._check_config(config_file=True, bases=True)

    with patch("charmcraft.main.COMMAND_GROUPS", [CommandGroup("title", [MyCommand])]):
        retcode = main(["charmcraft", "cmdname", f"--project-dir={tmp_path}"])
    assert retcode == 1

    out, err = capsys.readouterr()
    assert not out
    assert err.startswith(
        "The specified command needs a valid 'bases' in 'charmcraft.yaml' configuration "
        "file (in the current directory or where specified with --project-dir option)."
    )


def test_main_no_args():
    """The setup.py entry_point function needs to work with no arguments."""
    with patch("sys.argv", ["charmcraft"]):
        retcode = main()

    assert retcode == 1


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_main_controlled_error(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Work raised CraftError: message handler notified properly, use indicated return code."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    simulated_exception = CraftError("boom", retcode=33)
    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.side_effect = simulated_exception
            retcode = main(["charmcraft", "version"])

    assert retcode == 33
    emit_mock.error.assert_called_once_with(simulated_exception)


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        ["type: charm", None],
        [None, None],
        [
            dedent(
                """\
                name: test-charm-name-from-charmcraft-yaml
                type: charm
                bases:
                  - name: ubuntu
                    channel: "20.04
                """
            ),
            None,
        ],
        ["invalid file", "charmcraft shouldn't even read the files in this test."],
    ],
)
def test_main_controlled_return_code(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Work ended ok, and the command indicated the return code."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.return_value = 9
            retcode = main(["charmcraft", "version"])

    assert retcode == 9
    emit_mock.ended_ok.assert_called_once_with()


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        ["type: charm", None],
        [None, None],
        [
            dedent(
                """\
                name: test-charm-name-from-charmcraft-yaml
                type: charm
                bases:
                  - name: ubuntu
                    channel: "20.04
                """
            ),
            None,
        ],
        ["invalid file", "charmcraft shouldn't even read the files in this test."],
    ],
)
def test_main_crash(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Work crashed: message handler notified properly, return code in 1."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    simulated_exception = ValueError("boom")
    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.side_effect = simulated_exception
            retcode = main(["charmcraft", "version"])

    assert retcode == 1
    (call,) = emit_mock.error.mock_calls
    (exc,) = call.args
    assert isinstance(exc, CraftError)
    assert str(exc) == "charmcraft internal error: ValueError('boom')"
    assert exc.__cause__ == simulated_exception


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        ["type: charm", None],
        [None, None],
        [
            dedent(
                """\
                name: test-charm-name-from-charmcraft-yaml
                type: charm
                bases:
                  - name: ubuntu
                    channel: "20.04
                """
            ),
            None,
        ],
        ["invalid file", "charmcraft shouldn't even read the files in this test."],
    ],
)
def test_main_interrupted(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Work interrupted: message handler notified properly, return code in 1."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    simulated_exception = KeyboardInterrupt()
    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.side_effect = simulated_exception
            retcode = main(["charmcraft", "version"])

    assert retcode == 1
    (call,) = emit_mock.error.mock_calls
    (exc,) = call.args
    assert isinstance(exc, CraftError)
    assert str(exc) == "Interrupted."
    assert exc.__cause__ == simulated_exception


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        ["type: charm", None],
        [None, None],
        [
            dedent(
                """\
                name: test-charm-name-from-charmcraft-yaml
                type: charm
                bases:
                  - name: ubuntu
                    channel: "20.04
                """
            ),
            None,
        ],
        ["invalid file", "charmcraft shouldn't even read the files in this test."],
    ],
)
def test_main_controlled_arguments_error(
    capsys, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """The execution failed because an argument parsing error."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.side_effect = ArgumentParsingError("test error")
            retcode = main(["charmcraft", "version"])

    assert retcode == 1
    emit_mock.ended_ok.assert_called_once_with()

    out, err = capsys.readouterr()
    assert not out
    assert err == "test error\n"


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        ["type: charm", None],
        [None, None],
        [
            dedent(
                """\
                name: test-charm-name-from-charmcraft-yaml
                type: charm
                bases:
                  - name: ubuntu
                    channel: "20.04
                """
            ),
            None,
        ],
        ["invalid file", "charmcraft shouldn't even read the files in this test."],
    ],
)
def test_main_providing_help(
    capsys, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """The execution ended up providing a help message."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as d_mock:
            d_mock.side_effect = ProvideHelpException("nice and shiny help message")
            retcode = main(["charmcraft", "version"])

    assert retcode == 0
    emit_mock.ended_ok.assert_called_once_with()

    out, err = capsys.readouterr()
    assert not out
    assert err == "nice and shiny help message\n"


def test_main_logs_system_details(emitter, config):
    """Calling main ends up logging the system details."""
    system_details = "test system details"

    with patch("charmcraft.main.emit") as emit_mock:
        with patch("charmcraft.main.Dispatcher.run") as run_mock:
            with patch("charmcraft.main._get_system_details") as details_mock:
                details_mock.return_value = system_details
                run_mock.return_value = None
                main(["charmcraft", "version"])
    emit_mock.debug.assert_called_once_with(system_details)


# -- tests for system details producer


def test_systemdetails_basic():
    """Basic system details."""
    with patch("os.environ", {}):
        with patch("charmcraft.utils.get_os_platform") as platform_mock:
            platform_mock.return_value = utils.OSPlatform(
                system="test-system", release="test-release", machine="test-machine"
            )
            result = _get_system_details()
    assert result == (
        "System details: OSPlatform(system='test-system', release='test-release', "
        "machine='test-machine'); Environment: None"
    )


def test_systemdetails_extra_environment(monkeypatch):
    """System details with extra environment variables."""
    with patch("os.environ", {"TEST1": "test1", "TEST2": "test2", "TEST3": "test3"}):
        with patch("charmcraft.utils.get_os_platform") as platform_mock:
            platform_mock.return_value = utils.OSPlatform(
                system="test-system", release="test-release", machine="test-machine"
            )
            with patch("charmcraft.main.EXTRA_ENVIRONMENT", ("TEST1", "TEST3")):
                result = _get_system_details()
    assert result == (
        "System details: OSPlatform(system='test-system', release='test-release', "
        "machine='test-machine'); Environment: TEST1='test1', TEST3='test3'"
    )


def test_systemdetails_charmcraft_environment():
    """System details with environment variables specific to Charmcraft."""
    with patch("os.environ", {"CHARMCRAFT-TEST": "testvalue"}):
        with patch("charmcraft.utils.get_os_platform") as platform_mock:
            platform_mock.return_value = utils.OSPlatform(
                system="test-system", release="test-release", machine="test-machine"
            )
            result = _get_system_details()
    assert result == (
        "System details: OSPlatform(system='test-system', release='test-release', "
        "machine='test-machine'); Environment: CHARMCRAFT-TEST='testvalue'"
    )


def test_systemdetails_hidden_auth():
    """System details specifically hiding secrets."""
    with patch("os.environ", {ALTERNATE_AUTH_ENV_VAR: "supersecret"}):
        with patch("charmcraft.utils.get_os_platform") as platform_mock:
            platform_mock.return_value = utils.OSPlatform(
                system="test-system", release="test-release", machine="test-machine"
            )
            result = _get_system_details()
    assert result == (
        "System details: OSPlatform(system='test-system', release='test-release', "
        f"machine='test-machine'); Environment: {ALTERNATE_AUTH_ENV_VAR}='<hidden>'"
    )


# -- generic tests for all Charmcraft commands

all_commands = list(itertools.chain(*(cgroup.commands for cgroup in COMMAND_GROUPS)))


@pytest.mark.parametrize("command", all_commands)
def test_commands(command):
    """Assert commands are valid.

    This is done through asking help for it *in real life*, which would mean that the
    command is usable by the tool: that can be imported, instantiated, parse arguments, etc.
    """
    env = os.environ.copy()

    # Bypass unsupported environment error.
    env["CHARMCRAFT_DEVELOPER"] = "1"

    env_paths = [p for p in sys.path if "env/lib/python" in p]
    if env_paths:
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] += ":" + ":".join(env_paths)
        else:
            env["PYTHONPATH"] = ":".join(env_paths)

    external_command = [sys.executable, "-m", "charmcraft", command.name, "-h"]
    subprocess.run(external_command, check=True, env=env, stdout=subprocess.DEVNULL)


@pytest.mark.parametrize("command", all_commands)
def test_aesthetic_help_msg(command):
    """All real commands help msgs start with uppercase and do not end with a dot."""
    msg = command.help_msg
    assert msg[0].isupper()
    assert msg[-1] != "."


@pytest.mark.parametrize("command", all_commands)
def test_aesthetic_args_options_msg(command, config):
    """All real commands args help messages start with uppercase and do not end with a dot."""

    class FakeParser:
        """A fake to get the arguments added."""

        def add_mutually_exclusive_group(self, *args, **kwargs):
            """Return self, as it is used to add arguments too."""
            return self

        def add_argument(self, *args, **kwargs):
            """Verify that all commands have a correctly formatted help."""
            help_msg = kwargs.get("help")
            assert help_msg, "The help message must be present in each option"
            assert help_msg[0].isupper()
            assert help_msg[-1] != "."

    command(config).fill_parser(FakeParser())


@pytest.mark.parametrize("command_class", all_commands)
def test_usage_of_parsed_args(command_class, config):
    """The elements accessed on parsed_args need to be added before.

    This test is useful because normally all the tests for any command fake the
    Namespace and it happened in the past that we added functionality in the command
    execution but forgot to add the parameter in the 'fill_parser' method.
    """
    # get the list of attributes added by the command to the parser
    cmd = command_class(config)
    parser = ArgumentParser()
    cmd.fill_parser(parser)
    added_attributes = {action.dest for action in parser._actions}

    # build the abstract source tree for the command
    filepath = sys.modules[command_class.__module__].__file__
    tree = ast.parse(open(filepath).read())

    # get the node for the command
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef) and node.name == command_class.__name__:
            class_node = node
            break
    else:
        pytest.fail(f"Cannot find the class node for command {command_class}")

    # get the node for the 'run' function
    for node in ast.walk(class_node):
        if isinstance(node, ast.FunctionDef) and node.name == "run":
            run_method_node = node
            break
    else:
        pytest.fail(f"Cannot find the 'run' method node for command {command_class}")

    # get how the second argument to the function is called (normally
    # "parsed_args", but it's not enforced by the system)
    arg1, arg2 = run_method_node.args.args
    assert arg1.arg == "self"
    parsed_args_arg = arg2.arg

    for node in ast.walk(run_method_node):
        if not isinstance(node, ast.Attribute):
            continue
        attrib_value = node.value

        if isinstance(attrib_value, ast.Name) and attrib_value.id == parsed_args_arg:
            accessed_attribute = node.attr
            if accessed_attribute not in added_attributes:
                pytest.fail(
                    f"Found an accessed but not added argument ({accessed_attribute!r}) "
                    f"in command {command_class}"
                )


# -- tests for the base command


class MySimpleCommand(BaseCommand):
    """A simple minimal command to be used by different tests."""

    help_msg = "some help"
    name = "cmdname"
    overview = "test overview"


def test_basecommand_include_format_option(config):
    """Include a format option in the received parser."""
    parser = argparse.ArgumentParser()
    cmd = MySimpleCommand(config)
    cmd.include_format_option(parser)

    (action,) = (action for action in parser._actions if action.dest == "format")
    assert action.option_strings == ["--format"]
    assert action.dest == "format"
    assert action.default is None
    assert action.choices == [JSON_FORMAT]
    assert action.help == FORMAT_HELP_STR


def test_basecommand_format_content_json(config):
    """Properly format content in JSON format."""
    data = ["foo", "bar"]
    cmd = MySimpleCommand(config)
    result = cmd.format_content(JSON_FORMAT, data)
    assert result == dedent(
        """\
        [
            "foo",
            "bar"
        ]"""
    )


def test_basecommand_format_content_unkown(config):
    """The specified format is unknown."""
    cmd = MySimpleCommand(config)
    with pytest.raises(ValueError):
        cmd.format_content("bad format", {})



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

from textwrap import dedent
from typing import Any, Dict, List, Optional, Tuple

import pytest
from overrides import override

from charmcraft import errors, extensions
from charmcraft.config import load
from charmcraft.extensions.extension import Extension


class FakeExtension(Extension):
    """A fake test Extension"""

    name = "fake-extension"
    bases = [("ubuntu", "22.04")]

    @classmethod
    def get_supported_bases(cls) -> List[Tuple[str, ...]]:
        """Return a list of tuple of supported bases."""
        return cls.bases

    @staticmethod
    def is_experimental(_base: Optional[Tuple[str, ...]]) -> bool:
        """Return whether or not this extension is unstable for given base."""
        return False

    def get_root_snippet(self) -> Dict[str, Any]:
        """Return the root snippet to apply."""
        return {}

    def get_part_snippet(self) -> Dict[str, Any]:
        """Return the part snippet to apply to existing parts."""
        return {}

    def get_parts_snippet(self) -> Dict[str, Any]:
        """Return the parts to add to parts."""
        return {}


class ExperimentalExtension(FakeExtension):
    """A fake test Extension that is experimental"""

    name = "experimental-extension"
    bases = [("ubuntu", "22.04")]

    @staticmethod
    def is_experimental(_base: Optional[str]) -> bool:
        return True


class InvalidPartExtension(FakeExtension):
    """A fake test Extension that has invalid parts snippet"""

    name = "invalid-extension"
    bases = [("ubuntu", "22.04")]

    @override
    def get_parts_snippet(self) -> Dict[str, Any]:
        return {"bad-name": {"plugin": "dump", "source": None}}


class FullExtension(FakeExtension):
    """A fake test Extension that has complete behavior"""

    name = "full-extension"
    bases = [("ubuntu", "22.04")]

    @override
    def get_root_snippet(self) -> Dict[str, Any]:
        """Return the root snippet to apply."""
        return {
            "terms": ["https://example.com/terms", "https://example.com/terms2"],
        }

    @override
    def get_part_snippet(self) -> Dict[str, Any]:
        """Return the part snippet to apply to existing parts."""
        return {"stage-packages": ["new-package-1"]}

    @override
    def get_parts_snippet(self) -> Dict[str, Any]:
        """Return the parts to add to parts."""
        return {"full-extension/new-part": {"plugin": "nil", "source": None}}


@pytest.fixture()
def fake_extensions(stub_extensions):
    extensions.register(FakeExtension.name, FakeExtension)
    extensions.register(ExperimentalExtension.name, ExperimentalExtension)
    extensions.register(InvalidPartExtension.name, InvalidPartExtension)
    extensions.register(FullExtension.name, FullExtension)


def test_experimental_with_env(fake_extensions, tmp_path, monkeypatch):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [{"name": "ubuntu", "channel": "22.04"}],
        "extensions": [ExperimentalExtension.name],
    }
    monkeypatch.setenv("CHARMCRAFT_ENABLE_EXPERIMENTAL_EXTENSIONS", "1")
    project_root = tmp_path
    extensions.apply_extensions(project_root, charmcraft_config)


def test_experimental_no_env(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [
            {
                "build-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
                "run-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
            }
        ],
        "extensions": [ExperimentalExtension.name],
    }
    with pytest.raises(errors.ExtensionError) as exc:
        extensions.apply_extensions(tmp_path, charmcraft_config)

    expected_message = f"Extension is experimental: '{ExperimentalExtension.name}'"
    assert str(exc.value) == expected_message


def test_wrong_base(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [
            {
                "build-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
                "run-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
            }
        ],
        "extensions": [FakeExtension.name],
    }
    with pytest.raises(errors.ExtensionError) as exc:
        extensions.apply_extensions(tmp_path, charmcraft_config)

    expected_message = (
        f"Extension '{FakeExtension.name}' does not support base: ('ubuntu', '20.04')"
    )
    assert str(exc.value) == expected_message


def test_invalid_parts(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [{"name": "ubuntu", "channel": "22.04"}],
        "extensions": [InvalidPartExtension.name],
    }

    with pytest.raises(ValueError) as exc:
        extensions.apply_extensions(tmp_path, charmcraft_config)

    assert "Extension has invalid part names" in str(exc.value)


def test_apply_extensions(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [{"name": "ubuntu", "channel": "22.04"}],
        "extensions": [FullExtension.name],
        "parts": {"my-part": {"plugin": "nil", "source": None, "stage-packages": ["old-package"]}},
    }

    applied = extensions.apply_extensions(tmp_path, charmcraft_config)

    # Part snippet extends the existing part
    parts = applied["parts"]
    assert parts["my-part"]["stage-packages"] == [
        "new-package-1",
        "old-package",
    ]

    # New part
    assert parts[f"{FullExtension.name}/new-part"] == {"plugin": "nil", "source": None}


@pytest.mark.parametrize(
    ("charmcraft_yaml"),
    [
        dedent(
            f"""\
            type: charm
            name: test-charm-name-from-charmcraft-yaml
            summary: test summary
            description: test description
            bases:
              - name: ubuntu
                channel: "22.04"
            extensions:
              - {FullExtension.name}
            parts:
              foo:
                plugin: nil
                stage-packages:
                  - old-package
            """
        ),
    ],
)
def test_load_charmcraft_yaml_with_extensions(
    tmp_path,
    prepare_charmcraft_yaml,
    charmcraft_yaml,
    fake_extensions,
):
    """Load the config using charmcraft.yaml with extensions."""
    prepare_charmcraft_yaml(charmcraft_yaml)

    config = load(tmp_path)
    assert config.type == "charm"
    assert config.project.dirpath == tmp_path
    assert config.parts["foo"]["stage-packages"] == [
        "new-package-1",
        "old-package",
    ]

    # New part
    assert config.parts[f"{FullExtension.name}/new-part"] == {"plugin": "nil", "source": None}
    assert config.terms == ["https://example.com/terms", "https://example.com/terms2"]


@pytest.mark.parametrize(
    ("charmcraft_yaml"),
    [
        dedent(
            f"""\
            type: charm
            name: test-charm-name-from-charmcraft-yaml
            summary: test summary
            description: test description
            bases:
              - name: ubuntu
                channel: "20.04"
              - name: ubuntu
                channel: "22.04"
            extensions:
              - {FullExtension.name}
            parts:
              foo:
                plugin: nil
                stage-packages:
                  - old-package
            """
        ),
    ],
)
def test_load_charmcraft_yaml_with_extensions_unsupported_base(
    tmp_path,
    prepare_charmcraft_yaml,
    charmcraft_yaml,
    fake_extensions,
):
    """Load the config using charmcraft.yaml with extensions."""
    prepare_charmcraft_yaml(charmcraft_yaml)

    with pytest.raises(errors.ExtensionError) as exc:
        load(tmp_path)

    assert str(exc.value) == (
        "Extension 'full-extension' does not support base: ('ubuntu', '20.04')"
    )



# Copyright 2020-2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Tests for the Store client and authentication (code in store/client.py)."""

import base64
import json
from unittest import mock
from unittest.mock import call, patch

import craft_store
import pytest
import requests
from craft_cli import CraftError
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from charmcraft.commands.store.client import (
    AnonymousClient,
    Client,
    build_user_agent,
)
from charmcraft.utils import OSPlatform

# something well formed as tests exercise the internal machinery
ENCODED_CREDENTIALS = base64.b64encode(b"secret credentials").decode()


# --- General tests


def test_useragent_linux(monkeypatch):
    """Construct a user-agent as a patched Linux machine"""
    monkeypatch.setenv("TRAVIS_TESTING", "1")
    os_platform = OSPlatform(system="Arch Linux", release="5.10.10-arch1-1", machine="x86_64")
    with patch("charmcraft.commands.store.client.__version__", "1.2.3"), patch(
        "charmcraft.utils.get_os_platform", return_value=os_platform
    ), patch("platform.system", return_value="Linux"), patch(
        "platform.machine", return_value="x86_64"
    ), patch(
        "platform.python_version", return_value="3.9.1"
    ):
        ua = build_user_agent()
    assert ua == "charmcraft/1.2.3 (testing) Arch Linux/5.10.10-arch1-1 (x86_64) python/3.9.1"


def test_useragent_windows(monkeypatch):
    """Construct a user-agent as a patched Windows machine"""
    monkeypatch.setenv("TRAVIS_TESTING", "1")
    with patch("charmcraft.commands.store.client.__version__", "1.2.3"), patch(
        "platform.system", return_value="Windows"
    ), patch("platform.release", return_value="10"), patch(
        "platform.machine", return_value="AMD64"
    ), patch(
        "platform.python_version", return_value="3.9.1"
    ):
        ua = build_user_agent()
    assert ua == "charmcraft/1.2.3 (testing) Windows/10 (AMD64) python/3.9.1"


# --- Client tests


class FakeResponse(requests.Response):
    def __init__(self, content, status_code):
        self._content = content
        self.status_code = status_code

    @property
    def content(self):
        return self._content

    @property
    def ok(self):
        return self.status_code == 200

    def json(self):
        try:
            return json.loads(self._content)  # type: ignore
        except json.JSONDecodeError as exc:
            # the craft-store lib expects the error from requests, as what we're
            # faking here normally is a "real response"
            raise requests.exceptions.JSONDecodeError(exc.msg, exc.doc, exc.pos)

    @property
    def reason(self):
        return self._content

    @property
    def text(self):
        return self.content


@pytest.fixture()
def client_class():
    """Return a client instance with craft-store's StoreClient methods mocked."""

    auth_patch = patch("craft_store.Auth.__init__", return_value=None)
    auth_patch.start()

    class _StoreClientMock(craft_store.StoreClient):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            self.login_mock = mock.Mock()
            self.logout_mock = mock.Mock()
            self.request_mock = mock.Mock()

        def login(self, *args, **kwargs):
            self.login_mock(*args, **kwargs)

        def request(self, *args, **kwargs):
            return self.request_mock(*args, **kwargs)

        def logout(self, *args, **kwargs):
            self.logout_mock(*args, **kwargs)

    class _Client(Client, _StoreClientMock):
        pass

    yield _Client

    auth_patch.stop()


def test_client_init():
    """Check how craft-store's Client is initiated."""
    api_url = "http://api.test"
    storage_url = "http://storage.test"
    user_agent = "Super User Agent"
    with patch("craft_store.StoreClient.__init__") as mock_client_init:
        with patch("charmcraft.commands.store.client.build_user_agent") as mock_ua:
            mock_ua.return_value = user_agent
            Client(api_url, storage_url)
    mock_client_init.assert_called_with(
        base_url=api_url,
        storage_base_url=storage_url,
        endpoints=craft_store.endpoints.CHARMHUB,
        application_name="charmcraft",
        user_agent=user_agent,
        environment_auth="CHARMCRAFT_AUTH",
        ephemeral=False,
    )


def test_client_request_success_simple(client_class):
    """Hits the server, all ok."""
    response_value = {"foo": "bar"}
    fake_response = FakeResponse(content=json.dumps(response_value), status_code=200)
    client = client_class("http://api.test", "http://storage.test")
    client.request_mock.return_value = fake_response

    result = client.request_urlpath_json("GET", "/somepath")

    assert client.request_mock.mock_calls == [call("GET", "http://api.test/somepath")]
    assert result == response_value


def test_client_request_success_without_json_parsing(client_class):
    """Hits the server, all ok, return the raw response without parsing the json."""
    response_value = "whatever test response"
    fake_response = FakeResponse(content=response_value, status_code=200)
    client = client_class("http://api.test", "http://storage.test")
    client.request_mock.return_value = fake_response

    result = client.request_urlpath_text("GET", "/somepath")

    assert client.request_mock.mock_calls == [call("GET", "http://api.test/somepath")]
    assert result == response_value


def test_client_request_text_error(client_class):
    """Hits the server in text mode, getting an error."""
    client = client_class("http://api.test", "http://storage.test")
    original_error_text = "bad bad server"
    client.request_mock.side_effect = craft_store.errors.CraftStoreError(original_error_text)

    with pytest.raises(craft_store.errors.CraftStoreError) as cm:
        client.request_urlpath_text("GET", "/somepath")
    assert str(cm.value) == original_error_text


def test_client_request_json_error(client_class):
    """Hits the server in json mode, getting an error."""
    client = client_class("http://api.test", "http://storage.test")
    original_error_text = "bad bad server"
    client.request_mock.side_effect = craft_store.errors.CraftStoreError(original_error_text)

    with pytest.raises(craft_store.errors.CraftStoreError) as cm:
        client.request_urlpath_json("GET", "/somepath")
    assert str(cm.value) == original_error_text


def test_client_hit_success_withbody(client_class):
    """Hits the server including a body, all ok."""
    response_value = {"foo": "bar"}
    fake_response = FakeResponse(content=response_value, status_code=200)
    client = client_class("http://api.test", "http://storage.test")
    client.request_mock.return_value = fake_response

    result = client.request_urlpath_text("GET", "/somepath", "somebody")

    assert client.request_mock.mock_calls == [call("GET", "http://api.test/somepath", "somebody")]
    assert result == response_value


def test_client_init_removes_trailing_slashes(client_class):
    """The configured api url is ok even with an extra slash."""
    client = client_class("https://local.test:1234/", "http://storage.test/")

    assert client.api_base_url == "https://local.test:1234"
    assert client.storage_base_url == "http://storage.test"


def test_client_push_simple_ok(tmp_path, emitter, client_class):
    """Happy path for pushing bytes."""
    # fake some bytes to push
    test_filepath = tmp_path / "supercharm.bin"
    with test_filepath.open("wb") as fh:
        fh.write(b"abcdefgh")

    # used to store sizes from the monitor so we can assert emitter outside the fake
    # pusher after all is finished
    monitor_sizes_info = []

    def fake_pusher(monitor):
        """Push bytes in sequence, doing verifications in the middle."""
        # store the total monitor len (which is not only the saved bytes, but also
        # headers and stuff) and the prepared read batch size
        total_to_push = monitor.len
        read_size = int(total_to_push * 0.3)
        monitor_sizes_info.extend((total_to_push, read_size))

        # read twice the prepared chunk, and the rest
        monitor.read(read_size)
        monitor.read(read_size)
        monitor.read(total_to_push - read_size * 2)

        # check monitor is properly built
        assert isinstance(monitor.encoder, MultipartEncoder)
        filename, fh, ctype = monitor.encoder.fields["binary"]
        assert filename == "supercharm.bin"
        assert fh.name == str(test_filepath)
        assert ctype == "application/octet-stream"

        content = json.dumps({"successful": True, "upload_id": "test-upload-id"})

        return FakeResponse(content=content, status_code=200)

    client = client_class("http://api.test", "http://storage.test")
    with patch.object(client, "_storage_push", side_effect=fake_pusher):
        client.push_file(test_filepath)

    total_to_push, read_size = monitor_sizes_info
    emitter.assert_interactions(
        [
            call("progress", f"Starting to push {str(test_filepath)!r}"),
            call("progress_bar", "Uploading...", total_to_push, delta=False),
            call("advance", read_size),
            call("advance", read_size * 2),
            call("advance", total_to_push),
            call("progress", "Uploading bytes ended, id test-upload-id"),
        ]
    )


def test_client_push_configured_url_simple(tmp_path, client_class):
    """The storage server can be configured."""

    def fake_pusher(monitor):
        """Check the received URL."""
        content = json.dumps({"successful": True, "upload_id": "test-upload-id"})
        return FakeResponse(content=content, status_code=200)

    test_filepath = tmp_path / "supercharm.bin"
    test_filepath.write_text("abcdefgh")

    client = client_class("http://api.test", "https://local.test:1234/")
    with patch.object(client, "_storage_push", side_effect=fake_pusher):
        client.push_file(test_filepath)


def test_client_push_response_unsuccessful(tmp_path, client_class):
    """Didn't get a 200 from the Storage."""
    # fake some bytes to push
    test_filepath = tmp_path / "supercharm.bin"
    with test_filepath.open("wb") as fh:
        fh.write(b"abcdefgh")
    fake_response = FakeResponse(
        content=json.dumps({"successful": False, "upload_id": None}), status_code=200
    )

    client = client_class("http://api.test", "https://local.test:1234/")
    with patch.object(client, "_storage_push", return_value=fake_response):
        with pytest.raises(CraftError) as error:
            client.push_file(test_filepath)
            expected_error = (
                "Server error while pushing file: {'successful': False, 'upload_id': None}"
            )
            assert str(error.value) == expected_error


def test_storage_push_succesful(client_class):
    """Bytes are properly pushed to the Storage."""
    test_monitor = MultipartEncoderMonitor(
        MultipartEncoder(fields={"binary": ("filename", "somefile", "application/octet-stream")})
    )

    client = client_class("http://api.test", "http://test.url:0000")
    client._storage_push(test_monitor)

    # check request was properly called
    url = "http://test.url:0000/unscanned-upload/"
    headers = {
        "Content-Type": test_monitor.content_type,
        "Accept": "application/json",
    }
    assert client.request_mock.mock_calls == [
        call("POST", url, headers=headers, data=test_monitor)
    ]


def test_alternate_auth_login_forbidden(client_class, monkeypatch):
    """Login functionality cannot be used if alternate auth is present."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", ENCODED_CREDENTIALS)
    client = client_class("http://api.test", "http://storage.test")
    with pytest.raises(CraftError) as cm:
        client.login()
    expected_error = (
        "Cannot login when using alternative auth through CHARMCRAFT_AUTH environment variable."
    )
    assert str(cm.value) == expected_error


def test_alternate_auth_logout_forbidden(client_class, monkeypatch):
    """Logout functionality cannot be used if alternate auth is present."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", ENCODED_CREDENTIALS)
    client = client_class("http://api.test", "http://storage.test")
    with pytest.raises(CraftError) as cm:
        client.logout()
    expected_error = (
        "Cannot logout when using alternative auth through CHARMCRAFT_AUTH environment variable."
    )
    assert str(cm.value) == expected_error


def test_anonymous_client_init():
    """Check how craft-store's HTTPClient is initiated."""
    api_url = "http://api.test"
    storage_url = "http://storage.test"
    user_agent = "Super User Agent"
    with patch("craft_store.http_client.HTTPClient.__init__") as mock_client_init:
        with patch("charmcraft.commands.store.client.build_user_agent") as mock_ua:
            mock_ua.return_value = user_agent
            mock_client_init.return_value = None
            AnonymousClient(api_url, storage_url)

    mock_client_init.assert_called_with(
        user_agent=user_agent,
    )


def test_anonymous_client_request_success_simple():
    """Hits the server, all ok."""
    response_value = {"foo": "bar"}
    fake_response = FakeResponse(content=json.dumps(response_value), status_code=200)
    with patch("craft_store.http_client.HTTPClient.request") as mock_http_client_request:
        mock_http_client_request.return_value = fake_response
        client = AnonymousClient("http://api.test", "http://storage.test")
        result = client.request_urlpath_json("GET", "/somepath")

    assert mock_http_client_request.mock_calls == [call("GET", "http://api.test/somepath")]
    assert result == response_value


def test_anonymous_client_request_success_without_json_parsing():
    """Hits the server, all ok, return the raw response without parsing the json."""
    response_value = "whatever test response"
    fake_response = FakeResponse(content=response_value, status_code=200)
    with patch("craft_store.http_client.HTTPClient.request") as mock_http_client_request:
        client = AnonymousClient("http://api.test", "http://storage.test")
        mock_http_client_request.return_value = fake_response
        result = client.request_urlpath_text("GET", "/somepath")

    assert mock_http_client_request.mock_calls == [call("GET", "http://api.test/somepath")]
    assert result == response_value


def test_anonymous_client_request_text_error():
    """Hits the server in text mode, getting an error."""
    with patch("craft_store.http_client.HTTPClient.request") as mock_http_client_request:
        original_error_text = "bad bad server"
        mock_http_client_request.side_effect = craft_store.errors.CraftStoreError(
            original_error_text
        )
        client = AnonymousClient("http://api.test", "http://storage.test")

        with pytest.raises(craft_store.errors.CraftStoreError) as cm:
            client.request_urlpath_text("GET", "/somepath")

    assert str(cm.value) == original_error_text


def test_anonymous_client_request_json_error():
    """Hits the server in json mode, getting an error."""
    with patch("craft_store.http_client.HTTPClient.request") as mock_http_client_request:
        original_error_text = "bad bad server"
        mock_http_client_request.side_effect = craft_store.errors.CraftStoreError(
            original_error_text
        )
        client = AnonymousClient("http://api.test", "http://storage.test")

        with pytest.raises(craft_store.errors.CraftStoreError) as cm:
            client.request_urlpath_json("GET", "/somepath")

    assert str(cm.value) == original_error_text


def test_anonymous_client_hit_success_withbody():
    """Hits the server including a body, all ok."""
    response_value = {"foo": "bar"}
    fake_response = FakeResponse(content=response_value, status_code=200)
    with patch("craft_store.http_client.HTTPClient.request") as mock_http_client_request:
        mock_http_client_request.return_value = fake_response
        client = AnonymousClient("http://api.test", "http://storage.test")

        result = client.request_urlpath_text("GET", "/somepath", "somebody")

    assert mock_http_client_request.mock_calls == [
        call("GET", "http://api.test/somepath", "somebody")
    ]
    assert result == response_value


def test_anonymous_client_init_removes_trailing_slashes():
    """The configured api url is ok even with an extra slash."""
    client = AnonymousClient("https://local.test:1234/", "http://storage.test/")

    assert client.api_base_url == "https://local.test:1234"
    assert client.storage_base_url == "http://storage.test"



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import re
from unittest.mock import call, patch

import pytest

from charmcraft import deprecations
from charmcraft.deprecations import _DEPRECATION_MESSAGES, notify_deprecation


def test_notice_ok(monkeypatch, emitter):
    """Present proper messages to the user."""
    monkeypatch.setitem(_DEPRECATION_MESSAGES, "dn666", "Test message for the user.")
    monkeypatch.setattr(deprecations, "_DEPRECATION_URL_FMT", "http://docs.com/#{deprecation_id}")

    notify_deprecation("dn666")
    emitter.assert_interactions(
        [
            call("progress", "DEPRECATED: Test message for the user.", permanent=True),
            call("progress", "See http://docs.com/#dn666 for more information.", permanent=True),
        ]
    )


def test_notice_skipped_in_managed_mode(monkeypatch, emitter):
    """Present proper messages to the user."""
    monkeypatch.setitem(_DEPRECATION_MESSAGES, "dn666", "Test message for the user.")
    monkeypatch.setattr(deprecations, "_DEPRECATION_URL_FMT", "http://docs.com/#{deprecation_id}")

    with patch(
        "charmcraft.deprecations.is_charmcraft_running_in_managed_mode",
        return_value=True,
    ):
        notify_deprecation("dn666")

    emitter.assert_interactions(None)


@pytest.mark.parametrize("deprecation_id", _DEPRECATION_MESSAGES.keys())
def test_check_real_deprecation_ids(deprecation_id):
    """Verify all the real IDs used have the correct form."""
    assert re.match(r"dn\d\d", deprecation_id)


@pytest.mark.parametrize("message", _DEPRECATION_MESSAGES.values())
def test_check_real_deprecation_messages(message):
    """Verify all the real messages conform some rules."""
    assert message[0].isupper()
    assert message[-1] == "."


def test_log_deprecation_only_once(monkeypatch, emitter):
    """Show the message only once even if it was called several times."""
    monkeypatch.setitem(_DEPRECATION_MESSAGES, "dn666", "Test message for the user.")
    monkeypatch.setattr(deprecations, "_DEPRECATION_URL_FMT", "http://docs.com/#{deprecation_id}")

    # call twice, log once
    notify_deprecation("dn666")
    notify_deprecation("dn666")
    emitter.assert_interactions(
        [
            call("progress", "DEPRECATED: Test message for the user.", permanent=True),
            call("progress", "See http://docs.com/#dn666 for more information.", permanent=True),
        ]
    )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import os
from textwrap import dedent

import yaml

from charmcraft.config import load
from charmcraft.metafiles.actions import create_actions_yaml


def test_create_actions_yaml(tmp_path, prepare_charmcraft_yaml):
    """create actions.yaml."""
    actions = {
        "actions": {
            "pause": {"description": "Pause the database."},
            "resume": {"description": "Resume a paused database."},
            "snapshot": {
                "description": "Take a snapshot of the database.",
                "params": {
                    "filename": {
                        "type": "string",
                        "description": "The name of the snapshot file.",
                    },
                    "compression": {
                        "type": "object",
                        "description": "The type of compression to use.",
                        "properties": {
                            "kind": {"type": "string", "enum": ["gzip", "bzip2", "xz"]},
                            "quality": {
                                "description": "Compression quality",
                                "type": "integer",
                                "minimum": 0,
                                "maximum": 9,
                            },
                        },
                    },
                },
                "required": ["filename"],
                "additionalProperties": False,
            },
        }
    }

    yaml_data = yaml.safe_dump(actions)

    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description
            """
        )
        + yaml_data
    )

    config = load(tmp_path)

    actions_file = create_actions_yaml(tmp_path, config)

    assert yaml.safe_load(actions_file.read_text()) == actions["actions"]


def test_create_actions_yaml_none(tmp_path, prepare_charmcraft_yaml):
    """create actions.yaml with None, the file should not exist."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description
            """
        )
    )
    config = load(tmp_path)
    actions_file = create_actions_yaml(tmp_path, config)

    assert actions_file is None
    assert not os.path.exists(tmp_path / "actions.yaml")



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Infrastructure for the 'pack' command."""
import dataclasses
import os
import pathlib
import shutil
import subprocess
import tempfile
import zipfile
from typing import Collection, Dict, List, Mapping, Optional, Sequence

import craft_parts
import yaml
from craft_cli import CraftError, emit
from craft_providers.bases import get_base_alias

import charmcraft.env
import charmcraft.instrum
import charmcraft.linters
import charmcraft.providers
from charmcraft import const, env, errors, parts
from charmcraft.const import (
    BUILD_DIRNAME,
    CHARM_FILES,
    CHARM_OPTIONAL,
    UBUNTU_LTS_STABLE,
    VENV_DIRNAME,
)
from charmcraft.metafiles.actions import create_actions_yaml
from charmcraft.metafiles.config import create_config_yaml
from charmcraft.metafiles.manifest import create_manifest
from charmcraft.metafiles.metadata import create_metadata_yaml
from charmcraft.models.charmcraft import Base, BasesConfiguration
from charmcraft.models.lint import LintResult
from charmcraft.utils import (
    build_zip,
    collect_charmlib_pydeps,
    get_host_architecture,
    humanize_list,
    load_yaml,
)


@dataclasses.dataclass(frozen=True)
class OutputFiles:
    """Collection of output files, separated into charms and an optional bundle."""

    charms: Sequence[pathlib.Path] = ()
    bundles: Sequence[pathlib.Path] = ()


def _format_run_on_base(base: Base) -> str:
    """Formulate charm string for base section."""
    return "-".join([base.name, base.channel, *base.architectures])


def _format_bases_config(bases_config: BasesConfiguration) -> str:
    """Formulate charm string for bases configuration section."""
    return "_".join([_format_run_on_base(r) for r in bases_config.run_on])


def format_charm_file_name(charm_name: str, bases_config: BasesConfiguration) -> str:
    """Formulate charm file name.

    :param charm_name: Name of charm.
    :param bases_config: Bases configuration for charm.

    :returns: File name string, including .charm extension.
    """
    return "_".join([charm_name, _format_bases_config(bases_config)]) + ".charm"


def launch_shell(*, cwd: Optional[pathlib.Path] = None) -> None:
    """Launch a user shell for debugging environment.

    :param cwd: Working directory to start user in.
    """
    with emit.pause():
        subprocess.run(["bash"], check=False, cwd=cwd)


class Builder:
    """The package builder."""

    def __init__(
        self,
        *,
        config,
        force,
        debug,
        shell,
        shell_after,
        measure,
    ):
        self.force_packing = force
        self.debug = debug
        self.shell = shell
        self.shell_after = shell_after
        self.measure = measure

        self.charmdir = config.project.dirpath
        self.buildpath = self.charmdir / BUILD_DIRNAME
        self.shared_cache_path = charmcraft.env.get_host_shared_cache_path()

        self.config = config
        if self.config.parts:
            self._parts = self.config.parts.copy()
        else:
            self._parts = None

        # a part named "charm" using plugin "charm" is special and has
        # predefined values set automatically.
        charm_part = self._parts.get("charm")
        if charm_part and charm_part.get("plugin") == "charm":
            self._special_charm_part = charm_part
        else:
            self._special_charm_part = None

        self.provider = charmcraft.providers.get_provider()

    def show_linting_results(self, linting_results):
        """Manage the linters results, show some in different conditions, decide if continue."""
        attribute_results = []
        lint_results_by_outcome = {}
        for result in linting_results:
            if result.result == LintResult.IGNORED:
                continue
            if result.check_type == charmcraft.linters.CheckType.ATTRIBUTE:
                attribute_results.append(result)
            else:
                lint_results_by_outcome.setdefault(result.result, []).append(result)

        # show attribute results
        for result in attribute_results:
            emit.verbose(
                f"Check result: {result.name} [{result.check_type.value}] {result.result} "
                f"({result.text}; see more at {result.url}).",
            )

        # show warnings (if any), then errors (if any)
        template = "- {0.name}: {0.text} ({0.url})"
        if LintResult.WARNINGS in lint_results_by_outcome:
            emit.progress("Lint Warnings:", permanent=True)
            for result in lint_results_by_outcome[LintResult.WARNINGS]:
                emit.progress(template.format(result), permanent=True)
        if LintResult.ERRORS in lint_results_by_outcome:
            emit.progress("Lint Errors:", permanent=True)
            for result in lint_results_by_outcome[LintResult.ERRORS]:
                emit.progress(template.format(result), permanent=True)
            if self.force_packing:
                emit.progress("Packing anyway as requested.", permanent=True)
            else:
                raise CraftError(
                    "Aborting due to lint errors (use --force to override).", retcode=2
                )

    def build_charm(self, bases_config: BasesConfiguration) -> str:
        """Build the charm.

        :param bases_config: Bases configuration to use for build.

        :returns: File name of charm.

        :raises CraftError: on lifecycle exception.
        :raises RuntimeError: on unexpected lifecycle exception.
        """
        if charmcraft.env.is_charmcraft_running_in_managed_mode():
            work_dir = charmcraft.env.get_managed_environment_home_path()
        else:
            work_dir = self.buildpath

        emit.progress(f"Building charm in {str(work_dir)!r}")

        if self._special_charm_part:
            # add charm files to the prime filter
            # XXX Facundo 2022-07-18: we need to move this also to the plugin config
            self._set_prime_filter()

        # run the parts lifecycle
        emit.debug(f"Parts definition: {self._parts}")
        lifecycle = charmcraft.parts.PartsLifecycle(
            self._parts,
            work_dir=work_dir,
            project_dir=self.charmdir,
            project_name=self.config.name,
            ignore_local_sources=["*.charm"],
        )
        with charmcraft.instrum.Timer("Lifecycle run"):
            lifecycle.run(craft_parts.Step.PRIME)

        # skip creating yaml files if using reactive, reactive will create them
        # in a incompatible way
        if "reactive" not in {value.get("plugin") for value in self._parts.values()}:
            create_actions_yaml(lifecycle.prime_dir, self.config)
            create_config_yaml(lifecycle.prime_dir, self.config)
            create_metadata_yaml(lifecycle.prime_dir, self.config)

        # run linters and show the results
        linting_results = charmcraft.linters.analyze(self.config, lifecycle.prime_dir)
        self.show_linting_results(linting_results)

        create_manifest(
            lifecycle.prime_dir,
            self.config.project.started_at,
            bases_config,
            linting_results,
        )

        zipname = self.handle_package(lifecycle.prime_dir, bases_config)
        emit.progress(f"Created '{zipname}'.", permanent=True)
        return zipname

    def _set_prime_filter(self):
        """Add mandatory and optional charm files to the prime filter.

        The prime filter should contain:
        - The charm entry point, or the directory containing it if it's
          not directly in the project dir.
        - A set of mandatory charm files, including metadata.yaml, the
          dispatcher and the hooks directory.
        - A set of optional charm files.
        """
        charm_part_prime = self._special_charm_part.setdefault("prime", [])

        # add entrypoint
        entrypoint = pathlib.Path(self._special_charm_part["charm-entrypoint"])
        if str(entrypoint) == entrypoint.name:
            # the entry point is in the root of the project, just include it
            charm_part_prime.append(str(entrypoint))
        else:
            # the entry point is in a subdir, include the whole subtree
            charm_part_prime.append(str(entrypoint.parts[0]))

        # add venv if there are requirements
        charmlib_pydeps = collect_charmlib_pydeps(self.charmdir)
        if (
            self._special_charm_part.get("charm-requirements")
            or self._special_charm_part.get("charm-binary-python-packages")
            or self._special_charm_part.get("charm-python-packages")
            or charmlib_pydeps
        ):
            charm_part_prime.append(VENV_DIRNAME)

        # add mandatory and optional charm files
        charm_part_prime.extend(CHARM_FILES)
        for fn in CHARM_OPTIONAL:
            path = self.charmdir / fn
            if path.exists():
                charm_part_prime.append(fn)

    @charmcraft.instrum.Timer("Builder run")
    def run(
        self, bases_indices: Optional[List[int]] = None, destructive_mode: bool = False
    ) -> List[str]:
        """Run build process.

        In managed-mode or destructive-mode, build for each bases configuration
        which has a matching build-on to the host we are executing on.  Warn for
        each base configuration that is incompatible.  Error if unable to
        produce any builds for any bases configuration.

        :returns: List of charm files created.
        """
        charms: List[str] = []

        managed_mode = charmcraft.env.is_charmcraft_running_in_managed_mode()
        if not managed_mode and not destructive_mode:
            charmcraft.providers.ensure_provider_is_available(self.provider)

        build_plan = charmcraft.providers.create_build_plan(
            bases=self.config.bases,
            bases_indices=bases_indices,
            destructive_mode=destructive_mode,
            managed_mode=managed_mode,
            provider=self.provider,
        )
        if not build_plan:
            raise CraftError(
                "No suitable 'build-on' environment found in any 'bases' configuration."
            )

        charms = []
        for plan in build_plan:
            emit.debug(f"Building for 'bases[{plan.bases_index:d}][{plan.build_on_index:d}]'.")
            if managed_mode or destructive_mode:
                if self.shell:
                    # Execute shell in lieu of build.
                    launch_shell()
                    continue

                try:
                    with charmcraft.instrum.Timer("Building the charm"):
                        charm_name = self.build_charm(plan.bases_config)
                except (CraftError, RuntimeError) as error:
                    if self.debug:
                        emit.debug(f"Launching shell as charm building ended in error: {error}")
                        launch_shell()
                    raise

                if self.shell_after:
                    launch_shell()
            else:
                charm_name = self.pack_charm_in_instance(
                    bases_index=plan.bases_index,
                    build_on=plan.build_on,
                    build_on_index=plan.build_on_index,
                )
            charms.append(charm_name)

        return charms

    def pack_charm_in_instance(
        self, *, bases_index: int, build_on: Base, build_on_index: int
    ) -> str:
        """Pack instance in Charm."""
        charm_name = format_charm_file_name(self.config.name, self.config.bases[bases_index])

        # If building in project directory, use the project path as the working
        # directory. The output charms will be placed in the correct directory
        # without needing retrieval. If outputting to a directory other than the
        # charm project directory, we need to output the charm outside the
        # project directory and can retrieve it when complete.
        cwd = pathlib.Path.cwd()
        if cwd == self.charmdir:
            instance_output_dir = charmcraft.env.get_managed_environment_project_path()
            pull_charm = False
        else:
            instance_output_dir = charmcraft.env.get_managed_environment_home_path()
            pull_charm = True

        mode = emit.get_mode().name.lower()
        cmd = ["charmcraft", "pack", "--bases-index", str(bases_index), f"--verbosity={mode}"]

        if self.debug:
            cmd.append("--debug")

        if self.shell:
            cmd.append("--shell")

        if self.shell_after:
            cmd.append("--shell-after")

        if self.force_packing:
            cmd.append("--force")

        if self.measure:
            instance_metrics = charmcraft.env.get_managed_environment_metrics_path()
            cmd.append(f"--measure={str(instance_metrics)}")

        emit.progress(
            f"Launching environment to pack for base {build_on} "
            "(may take a while the first time but it's reusable)"
        )

        build_base_alias = get_base_alias((build_on.name, build_on.channel))
        instance_name = charmcraft.providers.get_instance_name(
            bases_index=bases_index,
            build_on_index=build_on_index,
            project_name=self.config.name,
            project_path=self.charmdir,
            target_arch=get_host_architecture(),
        )
        base_configuration = charmcraft.providers.get_base_configuration(
            alias=build_base_alias,
            instance_name=instance_name,
            shared_cache_path=self.shared_cache_path,
        )

        if build_on.name == "ubuntu":
            if build_on.channel in UBUNTU_LTS_STABLE:
                allow_unstable = False
            else:
                allow_unstable = True
                emit.message(
                    f"Warning: non-LTS Ubuntu releases {build_on.channel} are "
                    "intended for experimental use only."
                )
        else:
            allow_unstable = True
            emit.message(
                f"Warning: Base {build_on.name} {build_on.channel} daily image may be unstable."
            )

        with self.provider.launched_environment(
            project_name=self.config.name,
            project_path=self.charmdir,
            base_configuration=base_configuration,
            instance_name=instance_name,
            allow_unstable=allow_unstable,
        ) as instance:
            emit.debug("Mounting directory inside the instance")
            with charmcraft.instrum.Timer("Mounting directory"):
                instance.mount(
                    host_source=self.charmdir,
                    target=charmcraft.env.get_managed_environment_project_path(),
                )

            emit.progress("Packing the charm")
            emit.debug(f"Running {cmd}")
            try:
                with charmcraft.instrum.Timer("Execution inside instance"):
                    with emit.pause():
                        instance.execute_run(cmd, check=True, cwd=instance_output_dir)
                    if self.measure:
                        with instance.temporarily_pull_file(instance_metrics) as local_filepath:
                            charmcraft.instrum.merge_from(local_filepath)
            except subprocess.CalledProcessError as error:
                raise CraftError(
                    f"Failed to build charm for bases index '{bases_index}'."
                ) from error
            finally:
                charmcraft.providers.capture_logs_from_instance(instance)

            if pull_charm:
                try:
                    instance.pull_file(
                        source=instance_output_dir / charm_name,
                        destination=cwd / charm_name,
                    )
                except FileNotFoundError as error:
                    raise CraftError("Unexpected error retrieving charm from instance.") from error

        emit.progress("Charm packed ok")
        return charm_name

    def handle_package(self, prime_dir, bases_config: BasesConfiguration):
        """Handle the final package creation."""
        emit.progress("Creating the package itself")
        zipname = format_charm_file_name(self.config.name, bases_config)
        zipfh = zipfile.ZipFile(zipname, "w", zipfile.ZIP_DEFLATED)
        for dirpath, _dirnames, filenames in os.walk(prime_dir, followlinks=True):
            dirpath = pathlib.Path(dirpath)
            for filename in filenames:
                filepath = dirpath / filename
                zipfh.write(str(filepath), str(filepath.relative_to(prime_dir)))

        zipfh.close()
        return zipname

    def _get_charm_pack_args(self, base_indeces: List[str], destructive_mode: bool) -> List[str]:
        """Get the arguments for a charmcraft pack subprocess to run."""
        args = ["charmcraft", "pack", "--verbose"]
        if destructive_mode:
            args.append("--destructive-mode")
        for base in base_indeces:
            args.append(f"--bases-index={base}")
        if self.force_packing:
            args.append("--force")
        return args

    def pack_bundle(
        self,
        *,
        charms: Dict[str, pathlib.Path],
        base_indeces: List[str],
        destructive_mode: bool,
        overwrite: bool = False,
    ) -> OutputFiles:
        """Pack a bundle."""
        if self._parts is None:
            self._parts = {"bundle": {"plugin": "bundle"}}

        if env.is_charmcraft_running_in_managed_mode():
            work_dir = env.get_managed_environment_home_path()
        else:
            work_dir = self.config.project.dirpath / const.BUILD_DIRNAME

        # get the config files
        bundle_filepath = self.config.project.dirpath / "bundle.yaml"
        bundle = load_yaml(bundle_filepath)
        bundle_name = bundle.get("name")
        if not bundle_name:
            raise CraftError(
                "Invalid bundle config; missing a 'name' field indicating the bundle's name in "
                f"file {str(bundle_filepath)!r}."
            )

        if charms:
            bundle_charms = bundle.get("applications", {})
            command_args = self._get_charm_pack_args(base_indeces, destructive_mode)
            charms = _subprocess_pack_charms(charms, command_args)
            for name, value in bundle_charms.items():
                if name in charms:
                    value["charm"] = charms[name]
        else:
            charms = {}

        lifecycle = parts.PartsLifecycle(
            self._parts,
            work_dir=work_dir,
            project_dir=self.config.project.dirpath,
            project_name=bundle_name,
            ignore_local_sources=[bundle_name + ".zip"],
        )

        lifecycle.run(craft_parts.Step.PRIME)

        # pack everything
        create_manifest(
            lifecycle.prime_dir,
            self.config.project.started_at,
            bases_config=None,
            linting_results=[],
        )
        zipname = self.config.project.dirpath / (bundle_name + ".zip")
        if overwrite:
            primed_bundle_path = lifecycle.prime_dir / "bundle.yaml"
            with primed_bundle_path.open("w") as bundle_file:
                yaml.safe_dump(bundle, bundle_file)
        build_zip(zipname, lifecycle.prime_dir)

        return OutputFiles(charms=list(charms.values()), bundles=[zipname])


def _subprocess_pack_charms(
    charms: Mapping[str, pathlib.Path],
    command_args: Collection[str],
) -> Dict[str, pathlib.Path]:
    """Pack the given charms for a bundle in subprocesses.

    :param command_args: The initial arguments
    :param charms: A mapping of charm name to charm path
    :returns: A mapping of charm names to the generated charm.
    """
    if charms:
        charm_str = humanize_list(charms.keys(), "and")
        emit.progress(f"Packing charms: {charm_str}...")
    cwd = pathlib.Path(os.getcwd()).resolve()
    generated_charms = {}
    with tempfile.TemporaryDirectory(prefix="charmcraft-bundle-", dir=cwd) as temp_dir:
        temp_dir = pathlib.Path(temp_dir)
        try:
            # Put all the charms in this temporary directory.
            os.chdir(temp_dir)
            for charm, project_dir in charms.items():
                full_command = [*command_args, f"--project-dir={project_dir}"]
                with emit.open_stream(f"Packing charm {charm}...") as stream:
                    subprocess.check_call(full_command, stdout=stream, stderr=stream)
            duplicate_charms = {}
            for charm_file in temp_dir.glob("*.charm"):
                charm_name = charm_file.name.partition("_")[0]
                if charm_name not in charms:
                    emit.debug(f"Unknown charm file generated: {charm_file.name}")
                    continue
                if charm_name in generated_charms:
                    if charm_name not in duplicate_charms:
                        duplicate_charms[charm_name] = temp_dir.glob(f"{charm_name}_*.charm")
                    continue
                generated_charms[charm_name] = charm_file
            if duplicate_charms:
                raise errors.DuplicateCharmsError(duplicate_charms)
            for charm, charm_file in generated_charms.items():
                destination = cwd / charm_file.name
                destination.unlink(missing_ok=True)
                generated_charms[charm] = shutil.move(charm_file, destination)
        finally:
            os.chdir(cwd)
    return generated_charms



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
from textwrap import dedent
from unittest.mock import patch

import distro
import pytest

from charmcraft.utils.platform import OSPlatform, get_host_architecture, get_os_platform


@pytest.mark.parametrize(
    ("os_release", "expected_system", "expected_release"),
    [
        pytest.param(
            dedent(
                """
                # the following is an empty line

                NAME="Ubuntu"
                VERSION="20.04.1 LTS (Focal Fossa)"
                ID=ubuntu
                ID_LIKE=debian
                PRETTY_NAME="Ubuntu 20.04.1 LTS"
                VERSION_ID="20.04"
                HOME_URL="https://www.ubuntu.com/"
                SUPPORT_URL="https://help.ubuntu.com/"
                BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"

                # more in the middle; the following even would be "out of standard", but
                # we should not crash, just ignore it
                SOMETHING-WEIRD

                PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
                VERSION_CODENAME=focal
                UBUNTU_CODENAME=focal
                """
            ),
            "ubuntu",
            "20.04",
            id="ubuntu-focal",
        ),
        pytest.param(
            dedent(
                """
            NAME="Ubuntu Core"
            VERSION="20"
            ID=ubuntu-core
            PRETTY_NAME="Ubuntu Core 20"
            VERSION_ID="20"
            HOME_URL="https://snapcraft.io/"
            BUG_REPORT_URL="https://bugs.launchpad.net/snappy/"
            """
            ),
            "ubuntu-core",
            "20",
            id="ubuntu-core-20",
        ),
        pytest.param(
            dedent(
                """
                PRETTY_NAME="KDE neon 5.27"
                NAME="KDE neon"
                VERSION_ID="22.04"
                VERSION="5.27"
                VERSION_CODENAME=jammy
                ID=neon
                ID_LIKE="ubuntu debian"
                HOME_URL="https://neon.kde.org/"
                SUPPORT_URL="https://neon.kde.org/"
                BUG_REPORT_URL="https://bugs.kde.org/"
                PRIVACY_POLICY_URL="https://kde.org/privacypolicy/"
                UBUNTU_CODENAME=jammy
                LOGO=start-here-kde-neon
                """
            ),
            "ubuntu",
            "22.04",
            id="kde-neon-like-ubuntu",
        ),
    ],
)
@pytest.mark.parametrize("machine", ["x86_64", "riscv64", "arm64"])
def test_get_os_platform_linux(tmp_path, os_release, expected_system, expected_release, machine):
    """Utilize an /etc/os-release file to determine platform."""
    filepath = tmp_path / "os-release"
    filepath.write_text(os_release)
    with patch("distro.distro._distro", distro.LinuxDistribution(os_release_file=filepath)):
        with patch("platform.machine", return_value=machine):
            with patch("platform.system", return_value="Linux"):
                os_platform = get_os_platform(filepath)
    assert os_platform.system == expected_system
    assert os_platform.release == expected_release
    assert os_platform.machine == machine


@pytest.mark.parametrize("system", ["Windows", "Darwin", "Java", ""])
@pytest.mark.parametrize("release", ["NT", "Sparkling", "Jaguar", "0.0", ""])
@pytest.mark.parametrize("machine", ["AMD64", "x86_86", "arm64", "riscv64"])
def test_get_os_platform_non_linux(system, release, machine):
    """Get platform from a patched Windows machine."""
    with patch("platform.system", return_value=system):
        with patch("platform.release", return_value=release):
            with patch("platform.machine", return_value=machine):
                os_platform = get_os_platform()
    assert os_platform == OSPlatform(system, release, machine)


@pytest.mark.parametrize(
    ("platform_arch", "deb_arch"),
    [
        ("AMD64", "amd64"),
        ("aarch64", "arm64"),
        ("armv7l", "armhf"),
        ("ppc", "powerpc"),
        ("ppc64le", "ppc64el"),
        ("x86_64", "amd64"),
        ("unknown-arch", "unknown-arch"),
    ],
)
def test_get_host_architecture(platform_arch, deb_arch):
    """Test all platform mappings in addition to unknown."""
    with patch("platform.machine", return_value=platform_arch):
        assert get_host_architecture() == deb_arch
